import os
import sys

from hatchling.builders.hooks.plugin.interface import BuildHookInterface


def create_mo_files():
    from babel.messages.frontend import compile_catalog

    cmd = compile_catalog()
    cmd.directory = os.path.join(os.path.dirname(__file__), "canaille", "translations")
    cmd.quiet = True
    cmd.statistics = True
    cmd.finalize_options()
    return cmd.run()


class CustomBuildHook(BuildHookInterface):
    def initialize(self, version, build_data):
        create_mo_files()


if __name__ == "__main__":
    exit_code = create_mo_files()
    sys.exit(exit_code)



import datetime
import inspect
from pathlib import Path

from flask import current_app
from flask_alembic import Alembic
from sqlalchemy import create_engine
from sqlalchemy import or_
from sqlalchemy import select
from sqlalchemy import text
from sqlalchemy.orm import Session
from sqlalchemy.orm import declarative_base
from sqlalchemy_utils import Password

from canaille.app.configuration import CheckResult
from canaille.app.models import MODELS
from canaille.backends import Backend
from canaille.backends import ModelEncoder
from canaille.backends import get_lockout_delay_message
from canaille.backends.models import Model

Base = declarative_base()


class SQLModelEncoder(ModelEncoder):
    def default(self, obj):
        if isinstance(obj, Password):
            return obj.hash.decode()
        return super().default(obj)


class SQLBackend(Backend):
    engine = None
    db_session = None
    json_encoder = SQLModelEncoder
    alembic = None

    def __init__(self, config):
        super().__init__(config)
        SQLBackend.engine = create_engine(
            self.config["CANAILLE_SQL"]["DATABASE_URI"], echo=False, future=True
        )
        SQLBackend.alembic = Alembic(
            metadatas=Base.metadata, engines=SQLBackend.engine, run_mkdir=False
        )

    @classmethod
    def install(cls, app):  # pragma: no cover
        cls.init_alembic(app)
        SQLBackend.alembic.upgrade()

    @classmethod
    def init_alembic(cls, app):
        app.config["ALEMBIC"] = {
            "script_location": str(Path(__file__).resolve().parent / "migrations"),
        }
        SQLBackend.alembic.init_app(app)

    def init_app(self, app, init_backend=None):
        super().init_app(app)
        self.init_alembic(app)
        init_backend = (
            app.config["CANAILLE_SQL"]["AUTO_MIGRATE"]
            if init_backend is None
            else init_backend
        )
        if init_backend:  # pragma: no cover
            with app.app_context():
                self.alembic.upgrade()

    def setup(self):
        if not self.db_session:
            self.db_session = Session(SQLBackend.engine)

    def teardown(self):
        pass

    @classmethod
    def check_network_config(cls, config):
        sess = Session(SQLBackend.engine)
        sess.execute(text("SELECT 1"))
        return CheckResult(success=True, message="SQL database correctly configured")

    def has_account_lockability(self):
        return True

    def check_user_password(self, user, password):
        if current_app.features.has_intruder_lockout:
            if current_lockout_delay := user.get_intruder_lockout_delay():
                self.save(user)
                return (False, get_lockout_delay_message(current_lockout_delay))

        if password != user.password:
            if current_app.features.has_intruder_lockout:
                self.record_failed_attempt(user)
            return (False, None)

        if user.locked:
            return (False, "Your account has been locked.")

        return (True, None)

    def set_user_password(self, user, password):
        user.password = password
        user.password_last_update = datetime.datetime.now(
            datetime.timezone.utc
        ).replace(microsecond=0)
        self.save(user)

    def do_query(self, model, *args, **kwargs):
        filter = [
            model.attribute_filter(attribute_name, expected_value)
            for attribute_name, expected_value in kwargs.items()
        ]
        return (
            SQLBackend.instance.db_session.execute(select(model).filter(*filter))
            .scalars()
            .all()
        )

    def fuzzy(self, model, query, attributes=None, **kwargs):
        attributes = attributes or model.attributes
        filter = or_(
            getattr(model, attribute_name).ilike(f"%{query}%")
            for attribute_name in attributes
            if "str" in str(model.attributes[attribute_name])
            # erk, photo is an URL string according to SCIM, but bytes here
            and attribute_name != "photo"
        )

        return self.db_session.execute(select(model).filter(filter)).scalars().all()

    def get(self, model, identifier=None, /, **kwargs):
        if identifier:
            return (
                self.get(model, **{model.identifier_attribute: identifier})
                or self.get(model, id=identifier)
                or None
            )

        filter = [
            model.attribute_filter(attribute_name, expected_value)
            for attribute_name, expected_value in kwargs.items()
        ]
        return SQLBackend.instance.db_session.execute(
            select(model).filter(*filter)
        ).scalar_one_or_none()

    def do_restore(self, payload):
        # Create models without references to other models
        # We *could* do this in one single loop, but Client.audience is self-referencing,
        # so it has to be created first, and then must be added to its own audience.
        for model_name, model in MODELS.items():
            states = payload.get(model_name, [])
            for state in states:
                sql_state = self.filter_state(
                    model, state, keep_non_model_and_required_attrs=True
                )
                sql_state = self.replace_ids_with_instances(model, sql_state)
                obj = model(**sql_state)
                Backend.instance.save(obj)

        # Insert model references
        for model_name, model in MODELS.items():
            states = payload.get(model_name, [])
            for state in states:
                obj_id = state["id"]
                sql_state = self.filter_state(
                    model, state, keep_non_model_and_required_attrs=False
                )
                sql_state = self.replace_ids_with_instances(model, sql_state)
                obj = Backend.instance.get(model, obj_id)
                for attr, value in sql_state.items():
                    setattr(obj, attr, value)
                Backend.instance.save(obj)

    @classmethod
    def filter_state(cls, model, state, keep_non_model_and_required_attrs):
        def filter_attribute(attr):
            type_, _ = model.get_model_annotations(attr)
            is_non_model = not inspect.isclass(type_) or not issubclass(type_, Model)
            is_required = model.is_attr_required(attr)
            is_writable = not model.is_attr_readonly(attr)
            valid = (
                is_writable
                and (is_non_model or is_required) == keep_non_model_and_required_attrs
            )
            return valid

        return {attr: value for attr, value in state.items() if filter_attribute(attr)}

    @classmethod
    def replace_ids_with_instances(cls, model, state):
        for attr, value in state.items():
            generic_attr_model, _ = model.get_model_annotations(attr)
            if not generic_attr_model:
                continue

            model_name = generic_attr_model.__name__.lower()
            backend_attr_model = MODELS.get(model_name)
            if isinstance(value, list):
                state[attr] = [
                    Backend.instance.get(backend_attr_model, id_) for id_ in value
                ]
            else:
                state[attr] = Backend.instance.get(backend_attr_model, value)

        return state

    def do_save(self, instance):
        instance.last_modified = datetime.datetime.now(datetime.timezone.utc).replace(
            microsecond=0
        )
        if not instance.created:
            instance.created = instance.last_modified

        SQLBackend.instance.db_session.add(instance)
        SQLBackend.instance.db_session.commit()

    def do_delete(self, instance):
        SQLBackend.instance.db_session.delete(instance)
        SQLBackend.instance.db_session.commit()

    def do_reload(self, instance):
        SQLBackend.instance.db_session.refresh(instance)

    def record_failed_attempt(self, user):
        if user.password_failure_timestamps is None:
            user.password_failure_timestamps = []
        user._password_failure_timestamps.append(
            str(datetime.datetime.now(datetime.timezone.utc))
        )
        self.save(user)



import json
from unittest import mock

from canaille.commands import cli


def test_dump_stdout(testclient, backend, user, client, trusted_client, consent, token):
    """Test the full database dump command."""
    runner = testclient.app.test_cli_runner()
    res = runner.invoke(cli, ["dump"], catch_exceptions=False)
    assert res.exit_code == 0, res.stdout
    payload = json.loads(res.stdout)
    assert {
        "audience": [
            trusted_client.id,
        ],
        "client_id": mock.ANY,
        "client_id_issued_at": mock.ANY,
        "client_name": "Some other client",
        "client_secret": mock.ANY,
        "client_uri": "https://myotherdomain.test",
        "contacts": [
            "contact@myotherdomain.test",
        ],
        "created": mock.ANY,
        "grant_types": [
            "password",
            "authorization_code",
            "implicit",
            "hybrid",
            "refresh_token",
            "client_credentials",
            "urn:ietf:params:oauth:grant-type:jwt-bearer",
        ],
        "id": trusted_client.id,
        "jwks": mock.ANY,
        "last_modified": mock.ANY,
        "logo_uri": "https://myotherdomain.test/logo.webp",
        "policy_uri": "https://myotherdomain.test/policy",
        "post_logout_redirect_uris": [
            "https://myotherdomain.test/disconnected",
        ],
        "redirect_uris": [
            "https://myotherdomain.test/redirect1",
            "https://myotherdomain.test/redirect2",
        ],
        "response_types": [
            "code",
            "token",
            "id_token",
        ],
        "scope": [
            "openid",
            "profile",
            "groups",
        ],
        "token_endpoint_auth_method": "client_secret_basic",
        "tos_uri": "https://myotherdomain.test/tos",
        "trusted": True,
    } in payload["client"]
    assert {
        "audience": [
            client.id,
            trusted_client.id,
        ],
        "client_id": mock.ANY,
        "client_id_issued_at": mock.ANY,
        "client_name": "Some client",
        "client_secret": mock.ANY,
        "client_uri": "https://client.test",
        "contacts": [
            "contact@mydomain.test",
        ],
        "created": mock.ANY,
        "grant_types": [
            "password",
            "authorization_code",
            "implicit",
            "hybrid",
            "refresh_token",
            "client_credentials",
            "urn:ietf:params:oauth:grant-type:jwt-bearer",
        ],
        "id": client.id,
        "jwks": mock.ANY,
        "last_modified": mock.ANY,
        "logo_uri": "https://client.test/logo.webp",
        "policy_uri": "https://client.test/policy",
        "post_logout_redirect_uris": [
            "https://client.test/disconnected",
        ],
        "redirect_uris": [
            "https://client.test/redirect1",
            "https://client.test/redirect2",
        ],
        "response_types": [
            "code",
            "token",
            "id_token",
        ],
        "scope": [
            "openid",
            "email",
            "profile",
            "groups",
            "address",
            "phone",
        ],
        "token_endpoint_auth_method": "client_secret_basic",
        "tos_uri": "https://client.test/tos",
    } in payload["client"]
    assert payload["consent"] == [
        {
            "client": client.id,
            "consent_id": mock.ANY,
            "created": mock.ANY,
            "id": consent.id,
            "issue_date": mock.ANY,
            "last_modified": mock.ANY,
            "scope": [
                "openid",
                "profile",
            ],
            "subject": user.id,
        },
    ]
    assert payload["token"] == [
        {
            "access_token": mock.ANY,
            "audience": [
                client.id,
            ],
            "client": mock.ANY,
            "created": mock.ANY,
            "id": token.id,
            "issue_date": mock.ANY,
            "last_modified": mock.ANY,
            "lifetime": 3600,
            "refresh_token": mock.ANY,
            "scope": [
                "openid",
                "profile",
            ],
            "subject": user.id,
            "token_id": mock.ANY,
        },
    ]



def setup_security(app):
    try:
        from flask_talisman import Talisman
    except ImportError:
        return

    csp = {
        "default-src": "'self'",
        "font-src": "'self' data:",
    }
    Talisman(
        app,
        content_security_policy=csp,
        force_https=app.config["CANAILLE"]["FORCE_HTTPS"],
        session_cookie_secure=app.config["CANAILLE"]["FORCE_HTTPS"],
    )



def test_json_oauth_errors(testclient):
    """Checks that HTTP errors on the oauth endpoints are in the JSON format."""
    res = testclient.get("/oauth/invalid", status=404)
    assert res.json == {
        "error": "not_found",
        "error_description": "The requested URL was not found on the server. If you entered the URL manually please check your spelling and try again.",
    }



from flask import Blueprint
from flask import current_app
from flask import flash
from flask import request
from flask import url_for
from wtforms import StringField
from wtforms.validators import DataRequired

from canaille.app import build_hash
from canaille.app import obj_to_b64
from canaille.app.flask import user_needed
from canaille.app.forms import Form
from canaille.app.forms import email_validator
from canaille.app.i18n import gettext as _
from canaille.app.templating import render_template
from canaille.core.mails import send_test_mail

bp = Blueprint("admin", __name__, url_prefix="/admin")


class MailTestForm(Form):
    email = StringField(
        _("Email"),
        validators=[
            DataRequired(),
            email_validator,
        ],
        render_kw={
            "placeholder": _("jane.doe@example.com"),
            "spellcheck": "false",
            "autocorrect": "off",
        },
    )


@bp.route("/mail", methods=["GET", "POST"])
@user_needed("manage_oidc")
def mail_index(user):
    form = MailTestForm(request.form or None)
    if request.form and form.validate():
        if send_test_mail(form.email.data):
            flash(_("The test mail has been sent correctly"), "success")
        else:
            flash(_("The test mail has not been sent correctly"), "error")

    return render_template("core/mails/admin.html", form=form, menuitem="admin")


@bp.route("/mail/test.html")
@user_needed("manage_oidc")
def test_html(user):
    base_url = url_for("core.account.index", _external=True)
    return render_template(
        "core/mails/test.html",
        site_name=current_app.config["CANAILLE"]["NAME"],
        site_url=base_url,
        logo=current_app.config["CANAILLE"]["LOGO"],
        title=_("Test email from {website_name}").format(
            website_name=current_app.config["CANAILLE"]["NAME"],
        ),
    )


@bp.route("/mail/test.txt")
@user_needed("manage_oidc")
def test_txt(user):
    base_url = url_for("core.account.index", _external=True)
    return render_template(
        "core/mails/test.txt",
        site_name=current_app.config["CANAILLE"]["NAME"],
        site_url=current_app.config.get("SERVER_NAME", base_url),
    )


@bp.route("/mail/password-init.html")
@user_needed("manage_oidc")
def password_init_html(user):
    base_url = url_for("core.account.index", _external=True)
    reset_url = url_for(
        "core.auth.reset",
        user=user,
        token=user.generate_url_safe_token(),
        title=_("Password initialization on {website_name}").format(
            website_name=current_app.config["CANAILLE"]["NAME"]
        ),
        _external=True,
    )

    return render_template(
        "core/mails/firstlogin.html",
        site_name=current_app.config["CANAILLE"]["NAME"],
        site_url=base_url,
        reset_url=reset_url,
        logo=current_app.config["CANAILLE"]["LOGO"],
        title=_("Password initialization on {website_name}").format(
            website_name=current_app.config["CANAILLE"]["NAME"]
        ),
    )


@bp.route("/mail/password-init.txt")
@user_needed("manage_oidc")
def password_init_txt(user):
    base_url = url_for("core.account.index", _external=True)
    reset_url = url_for(
        "core.auth.reset",
        user=user,
        token=user.generate_url_safe_token(),
        _external=True,
    )

    return render_template(
        "core/mails/firstlogin.txt",
        site_name=current_app.config["CANAILLE"]["NAME"],
        site_url=current_app.config.get("SERVER_NAME", base_url),
        reset_url=reset_url,
    )


@bp.route("/mail/reset.html")
@user_needed("manage_oidc")
def password_reset_html(user):
    base_url = url_for("core.account.index", _external=True)
    server_name = current_app.config.get("SERVER_NAME")
    reset_token = user.generate_url_safe_token()
    reset_url = url_for(
        "core.auth.reset",
        user=user,
        token=reset_token,
        title=_("Password reset on {website_name}").format(
            website_name=current_app.config["CANAILLE"]["NAME"]
        ),
        _external=True,
    )

    return render_template(
        "core/mails/reset.html",
        site_name=current_app.config["CANAILLE"]["NAME"],
        site_url=base_url,
        reset_url=reset_url,
        server_name=server_name,
        reset_token=reset_token,
        reset_code=None,
        logo=current_app.config["CANAILLE"]["LOGO"],
        title=_("Password reset on {website_name}").format(
            website_name=current_app.config["CANAILLE"]["NAME"]
        ),
    )


@bp.route("/mail/reset.txt")
@user_needed("manage_oidc")
def password_reset_txt(user):
    base_url = url_for("core.account.index", _external=True)
    server_name = current_app.config.get("SERVER_NAME")
    reset_token = user.generate_url_safe_token()
    reset_url = url_for(
        "core.auth.reset",
        user=user,
        token=reset_token,
        _external=True,
    )

    return render_template(
        "core/mails/reset.txt",
        site_name=current_app.config["CANAILLE"]["NAME"],
        site_url=current_app.config.get("SERVER_NAME", base_url),
        reset_url=reset_url,
        server_name=server_name,
        reset_token=reset_token,
        reset_code=None,
    )


@bp.route("/mail/<identifier>/<email>/invitation.html")
@user_needed("manage_oidc")
def invitation_html(user, identifier, email):
    base_url = url_for("core.account.index", _external=True)
    registration_url = url_for(
        "core.account.registration",
        data=obj_to_b64([identifier, email]),
        hash=build_hash(identifier, email),
        _external=True,
    )

    return render_template(
        "core/mails/invitation.html",
        site_name=current_app.config["CANAILLE"]["NAME"],
        site_url=base_url,
        registration_url=registration_url,
        logo=current_app.config["CANAILLE"]["LOGO"],
        title=_("Invitation on {website_name}").format(
            website_name=current_app.config["CANAILLE"]["NAME"]
        ),
    )


@bp.route("/mail/<identifier>/<email>/invitation.txt")
@user_needed("manage_oidc")
def invitation_txt(user, identifier, email):
    base_url = url_for("core.account.index", _external=True)
    registration_url = url_for(
        "core.account.registration",
        data=obj_to_b64([identifier, email]),
        hash=build_hash(identifier, email),
        _external=True,
    )

    return render_template(
        "core/mails/invitation.txt",
        site_name=current_app.config["CANAILLE"]["NAME"],
        site_url=base_url,
        registration_url=registration_url,
    )


@bp.route("/mail/<identifier>/<email>/email-confirmation.html")
@user_needed("manage_oidc")
def email_confirmation_html(user, identifier, email):
    base_url = url_for("core.account.index", _external=True)
    email_confirmation_url = url_for(
        "core.account.email_confirmation",
        data=obj_to_b64([identifier, email]),
        hash=build_hash(identifier, email),
        _external=True,
    )

    return render_template(
        "core/mails/email-confirmation.html",
        site_name=current_app.config["CANAILLE"]["NAME"],
        site_url=base_url,
        confirmation_url=email_confirmation_url,
        logo=current_app.config["CANAILLE"]["LOGO"],
        title=_("Email confirmation on {website_name}").format(
            website_name=current_app.config["CANAILLE"]["NAME"]
        ),
    )


@bp.route("/mail/<identifier>/<email>/email-confirmation.txt")
@user_needed("manage_oidc")
def email_confirmation_txt(user, identifier, email):
    base_url = url_for("core.account.index", _external=True)
    email_confirmation_url = url_for(
        "core.account.email_confirmation",
        data=obj_to_b64([identifier, email]),
        hash=build_hash(identifier, email),
        _external=True,
    )

    return render_template(
        "core/mails/email-confirmation.txt",
        site_name=current_app.config["CANAILLE"]["NAME"],
        site_url=base_url,
        confirmation_url=email_confirmation_url,
    )


@bp.route("/mail/<email>/registration.html")
@user_needed("manage_oidc")
def registration_html(user, email):
    base_url = url_for("core.account.index", _external=True)
    registration_url = url_for(
        "core.account.registration",
        data=obj_to_b64([email]),
        hash=build_hash(email),
        _external=True,
    )

    return render_template(
        "core/mails/registration.html",
        site_name=current_app.config["CANAILLE"]["NAME"],
        site_url=base_url,
        registration_url=registration_url,
        logo=current_app.config["CANAILLE"]["LOGO"],
        title=_("Email confirmation on {website_name}").format(
            website_name=current_app.config["CANAILLE"]["NAME"]
        ),
    )


@bp.route("/mail/<email>/registration.txt")
@user_needed("manage_oidc")
def registration_txt(user, email):
    base_url = url_for("core.account.index", _external=True)
    registration_url = url_for(
        "core.account.registration",
        data=obj_to_b64([email]),
        hash=build_hash(email),
        _external=True,
    )

    return render_template(
        "core/mails/registration.txt",
        site_name=current_app.config["CANAILLE"]["NAME"],
        site_url=base_url,
        registration_url=registration_url,
    )


@bp.route("/mail/compromised_password_check_failure.html")
@user_needed("manage_oidc")
def compromised_password_check_failure_html(user):
    base_url = url_for("core.account.index", _external=True)
    user_name = "<USER NAME>"
    check_password_url = "<URL TO CHECK PASSWORD>"
    hashed_password = "<HASHED PASSWORD>"
    user_email = "<USER EMAIL>"

    return render_template(
        "core/mails/compromised_password_check_failure.html",
        site_name=current_app.config["CANAILLE"]["NAME"],
        site_url=base_url,
        user_name=user_name,
        check_password_url=check_password_url,
        hashed_password=hashed_password,
        logo=current_app.config["CANAILLE"]["LOGO"],
        user_email=user_email,
        title=_("Compromised password check failure on {website_name}").format(
            website_name=current_app.config["CANAILLE"]["NAME"]
        ),
    )


@bp.route("/mail/compromised_password_check_failure.txt")
@user_needed("manage_oidc")
def compromised_password_check_failure_txt(user):
    base_url = url_for("core.account.index", _external=True)
    user_name = "<USER NAME>"
    check_password_url = "<URL TO CHECK PASSWORD>"
    hashed_password = "<HASHED PASSWORD>"
    user_email = "<USER EMAIL>"

    return render_template(
        "core/mails/compromised_password_check_failure.txt",
        site_name=current_app.config["CANAILLE"]["NAME"],
        site_url=base_url,
        user_name=user_name,
        check_password_url=check_password_url,
        hashed_password=hashed_password,
        user_email=user_email,
    )


@bp.route("/mail/email_otp.html")
@user_needed("manage_oidc")
def email_otp_html(user):
    base_url = url_for("core.account.index", _external=True)
    otp = "000000"

    return render_template(
        "core/mails/email_otp.html",
        site_name=current_app.config["CANAILLE"]["NAME"],
        site_url=base_url,
        otp=otp,
        logo=current_app.config["CANAILLE"]["LOGO"],
        title=_("One-time password authentication on {website_name}").format(
            website_name=current_app.config["CANAILLE"]["NAME"]
        ),
    )


@bp.route("/mail/email_otp.txt")
@user_needed("manage_oidc")
def email_otp_txt(user):
    base_url = url_for("core.account.index", _external=True)
    otp = "000000"

    return render_template(
        "core/mails/email_otp.txt",
        site_name=current_app.config["CANAILLE"]["NAME"],
        site_url=base_url,
        otp=otp,
    )



import json
from http import HTTPStatus

from authlib.integrations.flask_oauth2 import ResourceProtector
from authlib.integrations.flask_oauth2.errors import (
    _HTTPException as AuthlibHTTPException,
)
from authlib.oauth2.rfc6750 import BearerTokenValidator
from flask import Blueprint
from flask import Response
from flask import abort
from flask import request
from pydantic import ValidationError
from scim2_models import Context
from scim2_models import EnterpriseUser
from scim2_models import Error
from scim2_models import ListResponse
from scim2_models import ResourceType
from scim2_models import Schema
from scim2_models import SearchRequest
from werkzeug.exceptions import HTTPException

from canaille.app import models
from canaille.app.flask import csrf
from canaille.backends import Backend

from .models import Group
from .models import User
from .models import get_resource_types
from .models import get_schemas
from .models import get_service_provider_config
from .models import group_from_canaille_to_scim_server
from .models import group_from_scim_to_canaille
from .models import user_from_canaille_to_scim_server
from .models import user_from_scim_to_canaille

bp = Blueprint("scim", __name__, url_prefix="/scim/v2")


class SCIMBearerTokenValidator(BearerTokenValidator):
    def authenticate_token(self, token_string: str):
        token = Backend.instance.get(models.Token, access_token=token_string)
        # At the moment, only client tokens are allowed, and not user tokens
        return token if token and not token.subject else None


require_oauth = ResourceProtector()
require_oauth.register_token_validator(SCIMBearerTokenValidator())


@bp.after_request
def add_scim_content_type(response):
    response.headers["Content-Type"] = "application/scim+json"
    return response


@bp.errorhandler(HTTPException)
def http_error_handler(error):
    obj = Error(detail=str(error), status=error.code)
    return obj.model_dump(), obj.status


@bp.errorhandler(AuthlibHTTPException)
def oauth2_error(error):
    body = json.loads(error.body)
    obj = Error(
        detail=f"{body['error']}: {body['error_description']}"
        if "error_description" in body
        else body["error"],
        status=error.code,
    )
    return obj.model_dump(), error.code


@bp.errorhandler(ValidationError)
def scim_error_handler(error):
    error_details = error.errors()[0]
    obj = Error(status=400, detail=error_details["msg"])
    # TODO: maybe the Pydantic <=> SCIM error code mapping could go in scim2_models
    obj.scim_type = (
        "invalidValue" if error_details["type"] == "required_error" else None
    )

    return obj.model_dump(), obj.status


def parse_search_request(request) -> SearchRequest:
    """Create a SearchRequest object from the request arguments."""
    max_nb_items_per_page = 1000
    count = (
        min(request.args["count"], max_nb_items_per_page)
        if request.args.get("count")
        else None
    )
    req = SearchRequest(
        attributes=request.args.get("attributes"),
        excluded_attributes=request.args.get("excludedAttributes"),
        start_index=request.args.get("startIndex"),
        count=count,
    )
    return req


@bp.route("/Users", methods=["GET"])
@csrf.exempt
@require_oauth()
def query_users():
    req = parse_search_request(request)
    users = list(
        Backend.instance.query(models.User)[req.start_index_0 : req.stop_index_0]
    )
    total = len(users)
    scim_users = [user_from_canaille_to_scim_server(user) for user in users]
    list_response = ListResponse[User[EnterpriseUser]](
        start_index=req.start_index,
        items_per_page=req.count,
        total_results=total,
        resources=scim_users,
    )
    payload = list_response.model_dump(
        scim_ctx=Context.RESOURCE_QUERY_RESPONSE,
    )
    return payload


@bp.route("/Users/<user:user>", methods=["GET"])
@csrf.exempt
@require_oauth()
def query_user(user):
    scim_user = user_from_canaille_to_scim_server(user)
    return scim_user.model_dump(
        scim_ctx=Context.RESOURCE_QUERY_RESPONSE,
    )


@bp.route("/Groups", methods=["GET"])
@csrf.exempt
@require_oauth()
def query_groups():
    req = parse_search_request(request)
    groups = list(
        Backend.instance.query(models.group)[req.start_index_0 : req.stop_index_0]
    )
    total = len(groups)
    scim_groups = [group_from_canaille_to_scim_server(group) for group in groups]
    list_response = ListResponse[Group](
        start_index=req.start_index,
        items_per_page=req.count,
        total_results=total,
        resources=scim_groups,
    )
    payload = list_response.model_dump(
        scim_ctx=Context.RESOURCE_QUERY_RESPONSE,
    )
    return payload


@bp.route("/Groups/<group:group>", methods=["GET"])
@csrf.exempt
@require_oauth()
def query_group(group):
    scim_group = group_from_canaille_to_scim_server(group)
    return scim_group.model_dump(
        scim_ctx=Context.RESOURCE_QUERY_RESPONSE,
    )


@bp.route("/Schemas", methods=["GET"])
@csrf.exempt
@require_oauth()
def query_schemas():
    req = parse_search_request(request)
    schemas = list(get_schemas().values())[req.start_index_0 : req.stop_index_0]
    response = ListResponse[Schema](
        total_results=len(schemas),
        items_per_page=req.count or len(schemas),
        start_index=req.start_index,
        resources=schemas,
    )
    return response.model_dump(scim_ctx=Context.RESOURCE_QUERY_RESPONSE)


@bp.route("/Schemas/<string:schema_id>", methods=["GET"])
@csrf.exempt
@require_oauth()
def query_schema(schema_id):
    schema = get_schemas().get(schema_id)
    if not schema:
        abort(404)

    return schema.model_dump(scim_ctx=Context.RESOURCE_QUERY_RESPONSE)


@bp.route("/ResourceTypes", methods=["GET"])
@csrf.exempt
@require_oauth()
def query_resource_types():
    req = parse_search_request(request)
    resource_types = list(get_resource_types().values())[
        req.start_index_0 : req.stop_index_0
    ]
    response = ListResponse[ResourceType](
        total_results=len(resource_types),
        items_per_page=req.count or len(resource_types),
        start_index=req.start_index,
        resources=resource_types,
    )
    return response.model_dump(scim_ctx=Context.RESOURCE_QUERY_RESPONSE)


@bp.route("/ResourceTypes/<string:resource_type_name>", methods=["GET"])
@csrf.exempt
@require_oauth()
def query_resource_type(resource_type_name):
    resource_type = get_resource_types().get(resource_type_name)
    if not resource_type:
        abort(404)

    return resource_type.model_dump(scim_ctx=Context.RESOURCE_QUERY_RESPONSE)


@bp.route("/ServiceProviderConfig", methods=["GET"])
@csrf.exempt
@require_oauth()
def query_service_provider_config():
    spc = get_service_provider_config()
    return spc.model_dump(scim_ctx=Context.RESOURCE_QUERY_RESPONSE)


@bp.route("/Users", methods=["POST"])
@csrf.exempt
@require_oauth()
def create_user():
    request_user = User[EnterpriseUser].model_validate(
        request.json, scim_ctx=Context.RESOURCE_CREATION_REQUEST
    )
    user = user_from_scim_to_canaille(request_user, models.User())
    Backend.instance.save(user)
    response_user = user_from_canaille_to_scim_server(user)
    payload = response_user.model_dump_json(scim_ctx=Context.RESOURCE_CREATION_RESPONSE)
    return Response(payload, status=HTTPStatus.CREATED)


@bp.route("/Groups", methods=["POST"])
@csrf.exempt
@require_oauth()
def create_group():
    request_group = Group.model_validate(
        request.json, scim_ctx=Context.RESOURCE_CREATION_REQUEST
    )
    group = group_from_scim_to_canaille(request_group, models.Group())
    Backend.instance.save(group)
    response_group = group_from_canaille_to_scim_server(group)
    payload = response_group.model_dump_json(
        scim_ctx=Context.RESOURCE_CREATION_RESPONSE
    )
    return Response(payload, status=HTTPStatus.CREATED)


@bp.route("/Users/<user:user>", methods=["PUT"])
@csrf.exempt
@require_oauth()
def replace_user(user):
    original_scim_user = user_from_canaille_to_scim_server(user)
    request_scim_user = User[EnterpriseUser].model_validate(
        request.json,
        scim_ctx=Context.RESOURCE_REPLACEMENT_REQUEST,
        original=original_scim_user,
    )
    updated_user = user_from_scim_to_canaille(request_scim_user, user)
    Backend.instance.save(updated_user)
    response_scim_user = user_from_canaille_to_scim_server(updated_user)
    payload = response_scim_user.model_dump(
        scim_ctx=Context.RESOURCE_REPLACEMENT_RESPONSE
    )
    return payload


@bp.route("/Groups/<group:group>", methods=["PUT"])
@csrf.exempt
@require_oauth()
def replace_group(group):
    original_scim_group = group_from_canaille_to_scim_server(group)
    request_scim_group = Group.model_validate(
        request.json,
        scim_ctx=Context.RESOURCE_REPLACEMENT_REQUEST,
        original=original_scim_group,
    )
    updated_group = group_from_scim_to_canaille(request_scim_group, group)
    Backend.instance.save(updated_group)
    response_group = group_from_canaille_to_scim_server(updated_group)
    payload = response_group.model_dump(scim_ctx=Context.RESOURCE_REPLACEMENT_RESPONSE)
    return payload


@bp.route("/Users/<user:user>", methods=["DELETE"])
@csrf.exempt
@require_oauth()
def delete_user(user):
    Backend.instance.delete(user)
    return "", HTTPStatus.NO_CONTENT


@bp.route("/Groups/<group:group>", methods=["DELETE"])
@csrf.exempt
@require_oauth()
def delete_group(group):
    Backend.instance.delete(group)
    return "", HTTPStatus.NO_CONTENT



"""Ordered relationships.

Revision ID: 1742033764
Revises: 1740765703
Create Date: 2025-03-15 11:16:04.256771

"""

from collections.abc import Sequence

import sqlalchemy as sa
from alembic import op

# revision identifiers, used by Alembic.
revision: str = "1742033764"
down_revision: str | None = "1740765703"
branch_labels: str | Sequence[str] | None = ()
depends_on: str | Sequence[str] | None = None


def upgrade() -> None:
    # ### commands auto generated by Alembic - please adjust! ###
    op.add_column(
        "client_audience_association_table",
        sa.Column("index", sa.Integer(), nullable=True),
    )
    op.execute("""
        UPDATE client_audience_association_table
        SET 'index' = s.row_num
        FROM (
            SELECT audience_id, client_id, ROW_NUMBER() OVER (ORDER BY audience_id, client_id) as row_num
            FROM client_audience_association_table
        ) s
        WHERE client_audience_association_table.audience_id = s.audience_id AND client_audience_association_table.client_id = s.client_id;
    """)
    with op.batch_alter_table("client_audience_association_table") as batch_op:
        batch_op.alter_column("index", existing_type=sa.Integer(), autoincrement=True)

    op.add_column(
        "membership_association_table",
        sa.Column("index", sa.Integer(), autoincrement=True, nullable=True),
    )
    op.execute("""
        UPDATE membership_association_table
        SET 'index' = s.row_num
        FROM (
            SELECT user_id, group_id, ROW_NUMBER() OVER (ORDER BY user_id, group_id) as row_num
            FROM membership_association_table
        ) s
        WHERE membership_association_table.user_id = s.user_id AND membership_association_table.group_id = s.group_id;
    """)
    with op.batch_alter_table("membership_association_table") as batch_op:
        batch_op.alter_column("index", existing_type=sa.Integer(), autoincrement=True)

    op.add_column(
        "token_audience_association_table",
        sa.Column("index", sa.Integer(), autoincrement=True, nullable=True),
    )
    op.execute("""
        UPDATE token_audience_association_table
        SET 'index' = s.row_num
        FROM (
            SELECT token_id, client_id, ROW_NUMBER() OVER (ORDER BY token_id, client_id) as row_num
            FROM token_audience_association_table
        ) s
        WHERE token_audience_association_table.token_id = s.token_id AND token_audience_association_table.client_id = s.client_id;
    """)
    with op.batch_alter_table("token_audience_association_table") as batch_op:
        batch_op.alter_column("index", existing_type=sa.Integer(), autoincrement=True)

    # ### end Alembic commands ###


def downgrade() -> None:
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_column("token_audience_association_table", "index")
    op.drop_column("membership_association_table", "index")
    op.drop_column("client_audience_association_table", "index")
    # ### end Alembic commands ###



import datetime
import json

from joserfc.jwk import KeySet
from joserfc.jwk import RSAKey
from werkzeug.security import gen_salt

from canaille.app import models


def test_no_logged_no_access(testclient):
    testclient.get("/admin/client", status=403)


def test_no_admin_no_access(testclient, logged_user):
    testclient.get("/admin/client", status=403)


def test_invalid_client_edition(testclient, logged_admin):
    testclient.get("/admin/client/edit/invalid", status=404)


def test_client_list(testclient, client, logged_admin):
    res = testclient.get("/admin/client")
    res.mustcontain(client.client_name)


def test_client_list_pagination(
    testclient, logged_admin, client, trusted_client, backend
):
    res = testclient.get("/admin/client")
    res.mustcontain("2 items")
    clients = []
    for _ in range(25):
        client = models.Client(client_id=gen_salt(48), client_name=gen_salt(48))
        backend.save(client)
        clients.append(client)

    res = testclient.get("/admin/client")
    res.mustcontain("27 items")
    client_name = res.pyquery(
        ".clients tbody tr:nth-of-type(1) td:nth-of-type(2) a"
    ).text()
    assert client_name

    form = res.forms["tableform"]
    res = form.submit(name="form", value="2")
    assert client_name not in res.pyquery(
        ".clients tbody tr td:nth-of-type(2) a"
    ).text().split(" ")
    for client in clients:
        backend.delete(client)

    res = testclient.get("/admin/client")
    res.mustcontain("2 items")


def test_client_list_bad_pages(testclient, logged_admin):
    res = testclient.get("/admin/client")
    form = res.forms["tableform"]
    testclient.post(
        "/admin/client",
        {"csrf_token": form["csrf_token"].value, "page": "2"},
        status=404,
    )

    res = testclient.get("/admin/client")
    form = res.forms["tableform"]
    testclient.post(
        "/admin/client",
        {"csrf_token": form["csrf_token"].value, "page": "-1"},
        status=404,
    )


def test_client_list_search(testclient, logged_admin, client, trusted_client):
    res = testclient.get("/admin/client")
    res.mustcontain("2 items")
    res.mustcontain(client.client_name)
    res.mustcontain(trusted_client.client_name)

    form = res.forms["search"]
    form["query"] = "other"
    res = form.submit()

    res.mustcontain("1 item")
    res.mustcontain(trusted_client.client_name)
    res.mustcontain(no=client.client_name)


def test_client_add(testclient, logged_admin, backend):
    res = testclient.get("/admin/client/add")
    data = {
        "client_name": "foobar",
        "contacts-0": "foo@bar.test",
        "client_uri": "https://foobar.test",
        "redirect_uris-0": "https://foobar.test/callback",
        "grant_types": ["password", "authorization_code"],
        "scope": "openid profile",
        "response_types": ["code", "token"],
        "token_endpoint_auth_method": "none",
        "logo_uri": "https://foobar.test/logo.webp",
        "tos_uri": "https://foobar.test/tos",
        "policy_uri": "https://foobar.test/policy",
        "software_id": "software",
        "software_version": "1",
        "jwks_uri": "https://foobar.test/jwks.json",
        "audience": [],
        "trusted": False,
        "post_logout_redirect_uris-0": "https://foobar.test/disconnected",
    }
    for k, v in data.items():
        res.form[k].force_value(v)

    res = res.form.submit(status=302, name="action", value="add")
    res = res.follow(status=200)

    client_id = res.forms["readonly"]["client_id"].value
    client = backend.get(models.Client, client_id=client_id)

    assert client.client_name == "foobar"
    assert client.contacts == ["foo@bar.test"]
    assert client.client_uri == "https://foobar.test"
    assert client.redirect_uris == ["https://foobar.test/callback"]
    assert client.grant_types == ["password", "authorization_code"]
    assert client.scope == ["openid", "profile"]
    assert client.response_types == ["code", "token"]
    assert client.token_endpoint_auth_method == "none"
    assert client.logo_uri == "https://foobar.test/logo.webp"
    assert client.tos_uri == "https://foobar.test/tos"
    assert client.policy_uri == "https://foobar.test/policy"
    assert client.software_id == "software"
    assert client.software_version == "1"
    assert client.jwks_uri == "https://foobar.test/jwks.json"
    assert client.audience == [client]
    assert not client.trusted
    assert client.post_logout_redirect_uris == ["https://foobar.test/disconnected"]

    backend.delete(client)


def test_add_missing_fields(testclient, logged_admin):
    res = testclient.get("/admin/client/add")
    res = res.form.submit(status=200, name="action", value="edit")
    assert (
        "error",
        "The client has not been added. Please check your information.",
    ) in res.flashes


def test_client_edit(testclient, client, logged_admin, trusted_client, backend):
    res = testclient.get("/admin/client/edit/" + client.client_id)
    data = {
        "client_name": "foobar",
        "contacts-0": "foo@bar.test",
        "client_uri": "https://foobar.test",
        "redirect_uris-0": "https://foobar.test/callback",
        "grant_types": ["password", "authorization_code"],
        "scope": "openid profile",
        "response_types": ["code", "token"],
        "token_endpoint_auth_method": "none",
        "logo_uri": "https://foobar.test/logo.webp",
        "tos_uri": "https://foobar.test/tos",
        "policy_uri": "https://foobar.test/policy",
        "software_id": "software",
        "software_version": "1",
        "jwks_uri": "https://foobar.test/jwks.json",
        "audience": [client.id, trusted_client.id],
        "trusted": True,
        "post_logout_redirect_uris-0": "https://foobar.test/disconnected",
    }
    for k, v in data.items():
        res.forms["clientaddform"][k].force_value(v)
    res = res.forms["clientaddform"].submit(status=302, name="action", value="edit")

    assert (
        "error",
        "The client has not been edited. Please check your information.",
    ) not in res.flashes
    assert ("success", "The client has been edited.") in res.flashes

    backend.reload(client)

    assert client.client_name == "foobar"
    assert client.contacts == ["foo@bar.test"]
    assert client.client_uri == "https://foobar.test"
    assert client.redirect_uris == [
        "https://foobar.test/callback",
        "https://client.test/redirect2",
    ]
    assert client.grant_types == ["password", "authorization_code"]
    assert client.scope == ["openid", "profile"]
    assert client.response_types == ["code", "token"]
    assert client.token_endpoint_auth_method == "none"
    assert client.logo_uri == "https://foobar.test/logo.webp"
    assert client.tos_uri == "https://foobar.test/tos"
    assert client.policy_uri == "https://foobar.test/policy"
    assert client.software_id == "software"
    assert client.software_version == "1"
    assert client.jwks_uri == "https://foobar.test/jwks.json"
    assert client.audience == [client, trusted_client]
    assert not client.trusted
    assert client.post_logout_redirect_uris == ["https://foobar.test/disconnected"]


def test_client_edit_missing_fields(
    testclient, client, logged_admin, trusted_client, backend
):
    res = testclient.get("/admin/client/edit/" + client.client_id)
    res.forms["clientaddform"]["client_name"] = ""
    res = res.forms["clientaddform"].submit(name="action", value="edit")
    assert (
        "error",
        "The client has not been edited. Please check your information.",
    ) in res.flashes
    backend.reload(client)
    assert client.client_name


def test_client_delete(testclient, logged_admin, backend):
    client = models.Client(client_id="client_id")
    backend.save(client)
    token = models.Token(
        token_id="id",
        client=client,
        subject=logged_admin,
        issue_date=datetime.datetime.now(datetime.timezone.utc),
    )
    backend.save(token)
    consent = models.Consent(
        consent_id="consent_id", subject=logged_admin, client=client, scope=["openid"]
    )
    backend.save(consent)
    authorization_code = models.AuthorizationCode(
        authorization_code_id="id", client=client, subject=logged_admin
    )
    backend.save(authorization_code)

    res = testclient.get("/admin/client/edit/" + client.client_id)
    res = res.forms["clientaddform"].submit(name="action", value="confirm-delete")
    res = res.form.submit(name="action", value="delete")
    res = res.follow()

    assert not backend.get(models.Client)
    assert not backend.get(models.Token)
    assert not backend.get(models.AuthorizationCode)
    assert not backend.get(models.Consent)


def test_client_delete_invalid_client(testclient, logged_admin, client):
    res = testclient.get(f"/admin/client/edit/{client.client_id}")
    testclient.post(
        "/admin/client/edit/invalid",
        {
            "action": "delete",
            "csrf_token": res.forms["clientaddform"]["csrf_token"].value,
        },
        status=404,
    )


def test_client_edit_preauth(testclient, client, logged_admin, trusted_client, backend):
    assert not client.trusted

    res = testclient.get("/admin/client/edit/" + client.client_id)
    res.forms["clientaddform"]["trusted"] = True
    res = res.forms["clientaddform"].submit(name="action", value="edit")

    assert ("success", "The client has been edited.") in res.flashes
    backend.reload(client)
    assert client.trusted

    res = testclient.get("/admin/client/edit/" + client.client_id)
    res.forms["clientaddform"]["trusted"] = False
    res = res.forms["clientaddform"].submit(name="action", value="edit")

    assert ("success", "The client has been edited.") in res.flashes
    backend.reload(client)
    assert not client.trusted


def test_client_edit_invalid_uri(testclient, client, logged_admin, trusted_client):
    res = testclient.get("/admin/client/edit/" + client.client_id)
    res.forms["clientaddform"]["client_uri"] = "invalid"
    res = res.forms["clientaddform"].submit(status=200, name="action", value="edit")
    assert (
        "error",
        "The client has not been edited. Please check your information.",
    ) in res.flashes
    res.mustcontain("This is not a valid URL")


def test_client_new_token(testclient, logged_admin, backend, client):
    res = testclient.get("/admin/client/edit/" + client.client_id)
    res = res.forms["clientaddform"].submit(name="action", value="new-token")
    assert (
        "success",
        "A token have been created for the client Some client",
    ) in res.flashes

    token = backend.get(models.Token)
    assert token.client == client
    assert not token.subject
    assert token.type == "access_token"
    assert token.scope == client.scope
    assert client in token.audience

    res = res.follow()
    assert res.template == "oidc/token_view.html"


def test_jwks_is_not_json(testclient, client, logged_admin, trusted_client, backend):
    res = testclient.get("/admin/client/edit/" + client.client_id)
    res.forms["clientaddform"]["jwks"] = "invalid"
    res = res.forms["clientaddform"].submit(status=200, name="action", value="edit")

    assert (
        "error",
        "The client has not been edited. Please check your information.",
    ) in res.flashes
    res.mustcontain("This value is not a valid JSON string.")


def test_jwks_is_not_jwks(testclient, client, logged_admin, trusted_client, backend):
    res = testclient.get("/admin/client/edit/" + client.client_id)
    res.forms["clientaddform"]["jwks"] = "{}"
    res = res.forms["clientaddform"].submit(status=200, name="action", value="edit")

    assert (
        "error",
        "The client has not been edited. Please check your information.",
    ) in res.flashes
    res.mustcontain("This value is not a valid JWK.")


def test_valid_jwk(testclient, client, logged_admin, trusted_client, backend):
    res = testclient.get("/admin/client/edit/" + client.client_id)
    keyset = KeySet([RSAKey.generate_key(1024)]).as_dict()
    res.forms["clientaddform"]["jwks"] = json.dumps(keyset)
    res = res.forms["clientaddform"].submit(status=302, name="action", value="edit")

    assert (
        "error",
        "The client has not been edited. Please check your information.",
    ) not in res.flashes
    assert ("success", "The client has been edited.") in res.flashes

    backend.reload(client)
    assert json.loads(client.jwks) == keyset



"""Methods related to HOTP and TOTP."""

import secrets

from blinker import signal
from flask import current_app

HOTP_LOOK_AHEAD_WINDOW = 10


def initialize_otp(user):
    user.secret_token = secrets.token_hex(32)
    user.last_otp_login = None
    if current_app.features.otp_method == "HOTP":
        user.hotp_counter = 1


def before_user_save(user, data):
    if current_app.features.has_otp and not user.secret_token:
        initialize_otp(user)


def generate_otp(user, counter_delta=0):
    import otpauth

    method = current_app.features.otp_method
    if method == "TOTP":
        totp = otpauth.TOTP(bytes(user.secret_token, "utf-8"))
        return totp.string_code(totp.generate())

    elif method == "HOTP":
        hotp = otpauth.HOTP(bytes(user.secret_token, "utf-8"))
        return hotp.string_code(hotp.generate(user.hotp_counter + counter_delta))

    else:  # pragma: no cover
        raise RuntimeError("Invalid one-time password method")


def get_otp_authentication_setup_uri(user):
    import otpauth

    method = current_app.features.otp_method
    if method == "TOTP":
        return otpauth.TOTP(bytes(user.secret_token, "utf-8")).to_uri(
            label=user.user_name, issuer=current_app.config["CANAILLE"]["NAME"]
        )

    elif method == "HOTP":
        return otpauth.HOTP(bytes(user.secret_token, "utf-8")).to_uri(
            label=user.user_name,
            issuer=current_app.config["CANAILLE"]["NAME"],
            counter=user.hotp_counter,
        )

    else:  # pragma: no cover
        raise RuntimeError("Invalid one-time password method")


def is_totp_valid(user, user_otp):
    import otpauth

    return otpauth.TOTP(bytes(user.secret_token, "utf-8")).verify(user_otp)


def is_hotp_valid(user, user_otp):
    import otpauth

    counter = user.hotp_counter
    is_valid = False
    # if user token's counter is ahead of canaille's, try to catch up to it
    while counter - user.hotp_counter <= HOTP_LOOK_AHEAD_WINDOW:
        is_valid = otpauth.HOTP(bytes(user.secret_token, "utf-8")).verify(
            user_otp, counter
        )
        counter += 1
        if is_valid:
            user.hotp_counter = counter
            return True
    return False


def setup_otp(app):
    signal("before_user_save").connect(before_user_save)



from authlib.oidc.core.grants.util import generate_id_token

from canaille.oidc.oauth import generate_user_info
from canaille.oidc.oauth import get_jwt_config


def test_end_session(testclient, backend, logged_user, client, id_token):
    testclient.get(f"/profile/{logged_user.user_name}", status=200)

    post_logout_redirect_url = "https://client.test/disconnected"
    res = testclient.get(
        "/oauth/end_session",
        params={
            "id_token_hint": id_token,
            "logout_hint": logged_user.user_name,
            "client_id": client.client_id,
            "post_logout_redirect_uri": post_logout_redirect_url,
            "state": "foobar",
        },
        status=302,
    )

    assert res.location == f"{post_logout_redirect_url}?state=foobar"

    with testclient.session_transaction() as sess:
        assert not sess.get("user_id")

    testclient.get(f"/profile/{logged_user.user_name}", status=403)


def test_end_session_no_client_id(testclient, backend, logged_user, client, id_token):
    testclient.get(f"/profile/{logged_user.user_name}", status=200)

    post_logout_redirect_url = "https://client.test/disconnected"
    res = testclient.get(
        "/oauth/end_session",
        params={
            "id_token_hint": id_token,
            "logout_hint": logged_user.user_name,
            "post_logout_redirect_uri": post_logout_redirect_url,
            "state": "foobar",
        },
        status=302,
    )

    assert res.location == f"{post_logout_redirect_url}?state=foobar"

    with testclient.session_transaction() as sess:
        assert not sess.get("user_id")

    testclient.get(f"/profile/{logged_user.user_name}", status=403)


def test_no_redirect_uri_no_redirect(
    testclient, backend, logged_user, client, id_token
):
    testclient.get(f"/profile/{logged_user.user_name}", status=200)

    res = testclient.get(
        "/oauth/end_session",
        params={
            "id_token_hint": id_token,
            "logout_hint": logged_user.user_name,
            "client_id": client.client_id,
            "state": "foobar",
        },
        status=302,
    )

    assert res.location == "/"

    with testclient.session_transaction() as sess:
        assert not sess.get("user_id")

    testclient.get(f"/profile/{logged_user.user_name}", status=403)


def test_bad_redirect_uri_no_redirect(
    testclient, backend, logged_user, client, id_token
):
    testclient.get(f"/profile/{logged_user.user_name}", status=200)

    post_logout_redirect_url = "https://client.test/invalid-uri"
    res = testclient.get(
        "/oauth/end_session",
        params={
            "id_token_hint": id_token,
            "logout_hint": logged_user.user_name,
            "client_id": client.client_id,
            "post_logout_redirect_uri": post_logout_redirect_url,
            "state": "foobar",
        },
        status=302,
    )

    assert res.location == "/"

    with testclient.session_transaction() as sess:
        assert not sess.get("user_id")

    testclient.get(f"/profile/{logged_user.user_name}", status=403)


def test_no_client_hint_no_redirect(testclient, backend, logged_user, client, id_token):
    testclient.get(f"/profile/{logged_user.user_name}", status=200)

    post_logout_redirect_url = "https://client.test/disconnected"
    res = testclient.get(
        "/oauth/end_session",
        params={
            "logout_hint": logged_user.user_name,
            "post_logout_redirect_uri": post_logout_redirect_url,
            "state": "foobar",
        },
        status=200,
    )
    res = res.form.submit(name="answer", value="logout", status=302)
    res = res.follow(status=302)

    assert res.location == "/"

    with testclient.session_transaction() as sess:
        assert not sess.get("user_id")

    testclient.get(f"/profile/{logged_user.user_name}", status=403)


def test_end_session_invalid_client_id(testclient, backend, logged_user, client):
    testclient.get(f"/profile/{logged_user.user_name}", status=200)

    post_logout_redirect_url = "https://client.test/disconnected"
    res = testclient.get(
        "/oauth/end_session",
        params={
            "logout_hint": logged_user.user_name,
            "post_logout_redirect_uri": post_logout_redirect_url,
            "client_id": "invalid_client_id",
            "state": "foobar",
        },
        status=200,
    )

    res = res.form.submit(name="answer", value="logout", status=302)
    res = res.follow(status=302)

    assert res.location == "/"

    with testclient.session_transaction() as sess:
        assert not sess.get("user_id")

    testclient.get(f"/profile/{logged_user.user_name}", status=403)


def test_client_hint_invalid(testclient, backend, logged_user, client):
    id_token = generate_id_token(
        {},
        generate_user_info(logged_user, client.scope),
        aud="invalid-client-id",
        **get_jwt_config(None),
    )

    testclient.get(f"/profile/{logged_user.user_name}", status=200)

    post_logout_redirect_url = "https://client.test/disconnected"
    res = testclient.get(
        "/oauth/end_session",
        params={
            "id_token_hint": id_token,
            "logout_hint": logged_user.user_name,
            "post_logout_redirect_uri": post_logout_redirect_url,
            "state": "foobar",
        },
        status=302,
    )

    assert res.location == "/"

    with testclient.session_transaction() as sess:
        assert not sess.get("user_id")

    testclient.get(f"/profile/{logged_user.user_name}", status=403)


def test_no_jwt_logout(testclient, backend, logged_user, client):
    testclient.get(f"/profile/{logged_user.user_name}", status=200)

    post_logout_redirect_url = "https://client.test/disconnected"
    res = testclient.get(
        "/oauth/end_session",
        params={
            "logout_hint": logged_user.user_name,
            "client_id": client.client_id,
            "post_logout_redirect_uri": post_logout_redirect_url,
            "state": "foobar",
        },
        status=200,
    )

    res = res.form.submit(name="answer", value="logout", status=302)
    res = res.follow(status=302)

    with testclient.session_transaction() as sess:
        assert not sess.get("user_id")

    assert res.location == f"{post_logout_redirect_url}?state=foobar"

    testclient.get(f"/profile/{logged_user.user_name}", status=403)


def test_no_jwt_no_logout(testclient, backend, logged_user, client):
    testclient.get(f"/profile/{logged_user.user_name}", status=200)

    post_logout_redirect_url = "https://client.test/disconnected"
    res = testclient.get(
        "/oauth/end_session",
        params={
            "logout_hint": logged_user.user_name,
            "client_id": client.client_id,
            "post_logout_redirect_uri": post_logout_redirect_url,
            "state": "foobar",
        },
        status=200,
    )

    res = res.form.submit(name="answer", value="stay", status=302)

    with testclient.session_transaction() as sess:
        assert sess.get("user_id")

    assert res.location == "/"

    testclient.get(f"/profile/{logged_user.user_name}", status=200)


def test_jwt_not_issued_here(testclient, backend, logged_user, client, id_token):
    testclient.app.config["CANAILLE_OIDC"]["JWT"]["ISS"] = "https://foobar.test"

    testclient.get(f"/profile/{logged_user.user_name}", status=200)

    post_logout_redirect_url = "https://client.test/disconnected"
    res = testclient.get(
        "/oauth/end_session",
        params={
            "id_token_hint": id_token,
            "logout_hint": logged_user.user_name,
            "client_id": client.client_id,
            "post_logout_redirect_uri": post_logout_redirect_url,
            "state": "foobar",
        },
        status=200,
    )

    assert res.json == {
        "message": "id_token_hint has not been issued here",
        "status": "error",
    }


def test_client_hint_mismatch(testclient, backend, logged_user, client):
    id_token = generate_id_token(
        {},
        generate_user_info(logged_user, client.scope),
        aud="another_client_id",
        **get_jwt_config(None),
    )

    testclient.get(f"/profile/{logged_user.user_name}", status=200)

    post_logout_redirect_url = "https://client.test/disconnected"
    res = testclient.get(
        "/oauth/end_session",
        params={
            "id_token_hint": id_token,
            "logout_hint": logged_user.user_name,
            "client_id": client.client_id,
            "post_logout_redirect_uri": post_logout_redirect_url,
            "state": "foobar",
        },
        status=200,
    )

    assert res.json == {
        "message": "id_token audience and client_id don't match",
        "status": "error",
    }


def test_end_session_bad_id_token(testclient, backend, logged_user, client, id_token):
    post_logout_redirect_url = "https://client.test/disconnected"
    res = testclient.get(
        "/oauth/end_session",
        params={
            "id_token_hint": "invalid",
            "logout_hint": logged_user.user_name,
            "client_id": client.client_id,
            "post_logout_redirect_uri": post_logout_redirect_url,
            "state": "foobar",
        },
    )

    assert res.json == {"status": "error", "message": "Invalid input segments length: "}


def test_bad_user_id_token_mismatch(testclient, backend, logged_user, client, admin):
    testclient.get(f"/profile/{logged_user.user_name}", status=200)

    id_token = generate_id_token(
        {},
        generate_user_info(admin, client.scope),
        aud=client.client_id,
        **get_jwt_config(None),
    )

    post_logout_redirect_url = "https://client.test/disconnected"
    res = testclient.get(
        "/oauth/end_session",
        params={
            "id_token_hint": id_token,
            "logout_hint": logged_user.user_name,
            "client_id": client.client_id,
            "post_logout_redirect_uri": post_logout_redirect_url,
            "state": "foobar",
        },
        status=200,
    )

    res = res.form.submit(name="answer", value="logout", status=302)
    res = res.follow(status=302)

    with testclient.session_transaction() as sess:
        assert not sess.get("user_id")

    assert res.location == f"{post_logout_redirect_url}?state=foobar"

    testclient.get(f"/profile/{logged_user.user_name}", status=403)


def test_bad_user_hint(testclient, backend, logged_user, client, id_token, admin):
    testclient.get(f"/profile/{logged_user.user_name}", status=200)

    post_logout_redirect_url = "https://client.test/disconnected"
    res = testclient.get(
        "/oauth/end_session",
        params={
            "id_token_hint": id_token,
            "logout_hint": admin.user_name,
            "client_id": client.client_id,
            "post_logout_redirect_uri": post_logout_redirect_url,
            "state": "foobar",
        },
        status=200,
    )

    res = res.form.submit(name="answer", value="logout", status=302)
    res = res.follow(status=302)

    with testclient.session_transaction() as sess:
        assert not sess.get("user_id")

    assert res.location == f"{post_logout_redirect_url}?state=foobar"

    testclient.get(f"/profile/{logged_user.user_name}", status=403)


def test_no_jwt_bad_csrf(testclient, backend, logged_user, client):
    testclient.get(f"/profile/{logged_user.user_name}", status=200)

    post_logout_redirect_url = "https://client.test/disconnected"
    res = testclient.get(
        "/oauth/end_session",
        params={
            "logout_hint": logged_user.user_name,
            "client_id": client.client_id,
            "post_logout_redirect_uri": post_logout_redirect_url,
            "state": "foobar",
        },
        status=200,
    )

    form = res.form
    form["csrf_token"] = "foobar"
    res = form.submit(name="answer", value="logout", status=400, expect_errors=True)


def test_end_session_already_disconnected(testclient, backend, user, client, id_token):
    post_logout_redirect_url = "https://client.test/disconnected"
    res = testclient.get(
        "/oauth/end_session",
        params={
            "id_token_hint": id_token,
            "logout_hint": user.user_name,
            "client_id": client.client_id,
            "post_logout_redirect_uri": post_logout_redirect_url,
            "state": "foobar",
        },
        status=302,
    )

    assert res.location == "/"


def test_end_session_no_state(testclient, backend, logged_user, client, id_token):
    testclient.get(f"/profile/{logged_user.user_name}", status=200)

    post_logout_redirect_url = "https://client.test/disconnected"
    res = testclient.get(
        "/oauth/end_session",
        params={
            "id_token_hint": id_token,
            "logout_hint": logged_user.user_name,
            "client_id": client.client_id,
            "post_logout_redirect_uri": post_logout_redirect_url,
        },
        status=302,
    )

    assert res.location == post_logout_redirect_url

    with testclient.session_transaction() as sess:
        assert not sess.get("user_id")

    testclient.get(f"/profile/{logged_user.user_name}", status=403)



import click
from flask.cli import with_appcontext

from canaille.app import models
from canaille.app.commands import with_backendcontext
from canaille.backends import Backend


@click.command()
@with_appcontext
@with_backendcontext
def clean():
    """Remove expired tokens and authorization codes."""
    for t in Backend.instance.query(models.Token):
        if t.is_expired():
            Backend.instance.delete(t)

    for a in Backend.instance.query(models.AuthorizationCode):
        if a.is_expired():
            Backend.instance.delete(a)


def register(cli):
    cli.add_command(clean)



def test_issuer(testclient, user):
    res = testclient.get(
        "/.well-known/webfinger?resource=acct%3Auser%40mydomain.test&rel=http%3A%2F%2Fopenid.net%2Fspecs%2Fconnect%2F1.0%2Fissuer"
    )
    assert res.json == {
        "subject": "acct:user@mydomain.test",
        "links": [
            {
                "rel": "http://openid.net/specs/connect/1.0/issuer",
                "href": "https://auth.test",
            }
        ],
    }


def test_resource_unknown(testclient):
    res = testclient.get(
        "/.well-known/webfinger?resource=acct%3Ainvalid%40mydomain.test&rel=http%3A%2F%2Fopenid.net%2Fspecs%2Fconnect%2F1.0%2Fissuer",
    )
    assert res.json == {
        "subject": "acct:invalid@mydomain.test",
        "links": [
            {
                "rel": "http://openid.net/specs/connect/1.0/issuer",
                "href": "https://auth.test",
            }
        ],
    }


def test_bad_request(testclient, user):
    testclient.get("/.well-known/webfinger", status=400)



import pytest

from canaille.app.configuration import settings_factory
from canaille.backends.sql.backend import SQLBackend


@pytest.fixture
def sqlalchemy_configuration(configuration):
    configuration["CANAILLE"]["DATABASE"] = "sql"
    configuration["CANAILLE_SQL"] = {"DATABASE_URI": "sqlite:///:memory:"}
    yield configuration
    del configuration["CANAILLE_SQL"]


@pytest.fixture
def sql_backend(sqlalchemy_configuration):
    config_obj = settings_factory(sqlalchemy_configuration)
    config_dict = config_obj.model_dump()
    backend = SQLBackend(config_dict)
    with backend.session():
        yield backend



import logging
from unittest import mock

from scim2_models import EnterpriseUser
from scim2_models import SearchRequest

from canaille.app import models
from canaille.scim.client import user_from_canaille_to_scim_client
from canaille.scim.models import User as MyUser


def test_scim_client_user_save_and_delete(scim_client_for_trusted_client, backend):
    User = scim_client_for_trusted_client.get_resource_model("User")

    response = scim_client_for_trusted_client.query(User)
    assert not response.resources

    alice = models.User(
        formatted_name="Alice Alice",
        family_name="Alice",
        user_name="alice",
        emails=["john@doe.test", "johhny@doe.test"],
    )
    backend.save(alice)

    response = scim_client_for_trusted_client.query(User)
    assert len(response.resources) == 1
    assert response.resources[0].user_name == "alice"

    backend.delete(alice)
    response = scim_client_for_trusted_client.query(User)
    assert not response.resources


def test_scim_client_group_save_and_delete(
    scim_client_for_trusted_client, backend, user
):
    Group = scim_client_for_trusted_client.get_resource_model("Group")
    User = scim_client_for_trusted_client.get_resource_model("User")

    req = SearchRequest(filter=f'externalId eq "{user.id}"')
    response = scim_client_for_trusted_client.query(User, search_request=req)
    distant_scim_user = response.resources[0] if response.resources else None

    response = scim_client_for_trusted_client.query(Group)
    assert not response.resources

    group = models.Group(
        members=[user],
        display_name="foobar",
    )
    backend.save(group)

    response = scim_client_for_trusted_client.query(Group)
    assert len(response.resources) == 1
    retrieved_group = response.resources[0]
    assert retrieved_group.display_name == "foobar"
    assert retrieved_group.members[0].value == distant_scim_user.id

    backend.delete(group)
    response = scim_client_for_trusted_client.query(Group)
    assert not response.resources


def test_scim_client_group_save_unable_to_retrieve_member_via_scim(
    scim_client_for_trusted_client, backend, user, caplog
):
    Group = scim_client_for_trusted_client.get_resource_model("Group")
    User = scim_client_for_trusted_client.get_resource_model("User")

    req = SearchRequest(filter=f'externalId eq "{user.id}"')
    response = scim_client_for_trusted_client.query(User, search_request=req)
    distant_scim_user = response.resources[0] if response.resources else None

    response = scim_client_for_trusted_client.query(Group)
    assert not response.resources

    group = models.Group(
        members=[user],
        display_name="foobar",
    )
    backend.save(group)

    scim_client_for_trusted_client.delete(User, distant_scim_user.id)

    backend.save(group)

    assert (
        "canaille",
        logging.WARNING,
        "Unable to find user user from group foobar via SCIM",
    ) in caplog.record_tuples

    backend.delete(group)


def test_scim_client_change_user_groups_also_updates_group_members(
    testclient,
    scim_client_for_trusted_client,
    backend,
    user,
    logged_admin,
    bar_group,
):
    Group = scim_client_for_trusted_client.get_resource_model("Group")
    User = scim_client_for_trusted_client.get_resource_model("User")

    req = SearchRequest(filter=f'externalId eq "{user.id}"')
    response = scim_client_for_trusted_client.query(User, search_request=req)
    distant_scim_user = response.resources[0] if response.resources else None

    req = SearchRequest(filter=f'externalId eq "{logged_admin.id}"')
    response = scim_client_for_trusted_client.query(User, search_request=req)
    distant_scim_admin = response.resources[0] if response.resources else None

    user.groups = user.groups + [bar_group]
    backend.save(user)

    response = scim_client_for_trusted_client.query(Group)
    assert len(response.resources) == 1
    retrieved_bar_group = response.resources[0]
    assert retrieved_bar_group.display_name == "bar"
    assert len(retrieved_bar_group.members) == 2
    assert retrieved_bar_group.members[0].value == distant_scim_admin.id
    assert retrieved_bar_group.members[1].value == distant_scim_user.id

    user.groups = []
    backend.save(user)

    response = scim_client_for_trusted_client.query(Group)
    assert len(response.resources) == 1
    retrieved_bar_group = response.resources[0]
    assert retrieved_bar_group.display_name == "bar"
    assert len(retrieved_bar_group.members) == 1
    assert retrieved_bar_group.members[0].value == distant_scim_admin.id


def test_scim_client_user_creation_and_deletion_also_updates_their_groups(
    testclient,
    scim_client_for_trusted_client,
    backend,
    foo_group,
    bar_group,
    user,
    admin,
    logged_admin,
):
    testclient.app.config["CANAILLE"]["ENABLE_REGISTRATION"] = True
    testclient.app.config["CANAILLE"]["EMAIL_CONFIRMATION"] = False

    Group = scim_client_for_trusted_client.get_resource_model("Group")
    User = scim_client_for_trusted_client.get_resource_model("User")

    req = SearchRequest(filter=f'externalId eq "{user.id}"')
    response = scim_client_for_trusted_client.query(User, search_request=req)
    distant_scim_user = response.resources[0] if response.resources else None

    req = SearchRequest(filter=f'externalId eq "{admin.id}"')
    response = scim_client_for_trusted_client.query(User, search_request=req)
    distant_scim_admin = response.resources[0] if response.resources else None

    response = scim_client_for_trusted_client.query(Group)
    assert len(response.resources) == 2
    retrieved_foo_group = response.resources[0]
    assert retrieved_foo_group.display_name == "foo"
    assert len(retrieved_foo_group.members) == 1
    assert retrieved_foo_group.members[0].value == distant_scim_user.id
    retrieved_bar_group = response.resources[1]
    assert retrieved_bar_group.display_name == "bar"
    assert len(retrieved_bar_group.members) == 1
    assert retrieved_bar_group.members[0].value == distant_scim_admin.id

    alice = models.User(
        formatted_name="Alice Alice",
        family_name="Alice",
        user_name="alice",
        emails=["john@doe.test", "johhny@doe.test"],
    )
    alice.groups = [bar_group, foo_group]
    backend.save(alice)

    req = SearchRequest(filter=f'externalId eq "{alice.id}"')
    response = scim_client_for_trusted_client.query(User, search_request=req)
    distant_scim_alice = response.resources[0] if response.resources else None

    response = scim_client_for_trusted_client.query(Group)
    assert len(response.resources) == 2
    retrieved_foo_group = response.resources[0]
    assert retrieved_foo_group.display_name == "foo"
    assert len(retrieved_foo_group.members) == 2
    assert retrieved_foo_group.members[0].value == distant_scim_user.id
    assert retrieved_foo_group.members[1].value == distant_scim_alice.id
    retrieved_bar_group = response.resources[1]
    assert retrieved_bar_group.display_name == "bar"
    assert len(retrieved_bar_group.members) == 2
    assert retrieved_bar_group.members[0].value == distant_scim_admin.id
    assert retrieved_bar_group.members[1].value == distant_scim_alice.id

    backend.delete(alice)

    response = scim_client_for_trusted_client.query(Group)
    assert len(response.resources) == 2
    retrieved_foo_group = response.resources[0]
    assert retrieved_foo_group.display_name == "foo"
    assert len(retrieved_foo_group.members) == 1
    assert retrieved_foo_group.members[0].value == distant_scim_user.id
    retrieved_bar_group = response.resources[1]
    assert retrieved_bar_group.display_name == "bar"
    assert len(retrieved_bar_group.members) == 1
    assert retrieved_bar_group.members[0].value == distant_scim_admin.id


def test_save_user_when_client_doesnt_support_scim(backend, user, consent, caplog):
    backend.save(user)
    assert (
        "canaille",
        logging.INFO,
        "SCIM protocol not supported by client Client",
    ) in caplog.record_tuples


def test_save_group_when_client_doesnt_support_scim(
    backend, bar_group, consent, caplog
):
    backend.save(bar_group)
    assert (
        "canaille",
        logging.INFO,
        "SCIM protocol not supported by client Client",
    ) in caplog.record_tuples


@mock.patch("scim2_client.engines.httpx.SyncSCIMClient.create")
def test_failed_scim_user_creation(
    scim_mock,
    testclient,
    scim_client_for_trusted_client,
    backend,
    caplog,
):
    scim_mock.side_effect = mock.Mock(side_effect=Exception())

    alice = models.User(
        formatted_name="Alice Alice",
        family_name="Alice",
        user_name="alice",
        emails=["john@doe.test", "johhny@doe.test"],
    )
    backend.save(alice)

    assert (
        "canaille",
        logging.WARNING,
        "SCIM User alice creation for client Some client failed",
    ) in caplog.record_tuples

    backend.delete(alice)


@mock.patch("scim2_client.engines.httpx.SyncSCIMClient.replace")
def test_failed_scim_user_update(
    scim_mock,
    testclient,
    scim_client_for_trusted_client,
    backend,
    caplog,
    user,
):
    scim_mock.side_effect = mock.Mock(side_effect=Exception())

    backend.save(user)

    assert (
        "canaille",
        logging.WARNING,
        "SCIM User user update for client Some client failed",
    ) in caplog.record_tuples


@mock.patch("scim2_client.engines.httpx.SyncSCIMClient.delete")
def test_failed_scim_user_delete(
    scim_mock,
    testclient,
    scim_client_for_trusted_client,
    backend,
    caplog,
    user,
):
    scim_mock.side_effect = mock.Mock(side_effect=Exception())

    backend.delete(user)

    assert (
        "canaille",
        logging.WARNING,
        "SCIM User user delete for client Some client failed",
    ) in caplog.record_tuples


@mock.patch("scim2_client.engines.httpx.SyncSCIMClient.create")
def test_failed_scim_group_creation(
    scim_mock,
    testclient,
    scim_client_for_trusted_client,
    backend,
    caplog,
    user,
):
    scim_mock.side_effect = mock.Mock(side_effect=Exception())

    group = models.Group(
        members=[user],
        display_name="foobar",
    )
    backend.save(group)

    assert (
        "canaille",
        logging.WARNING,
        "SCIM Group foobar creation for client Some client failed",
    ) in caplog.record_tuples

    backend.delete(group)


@mock.patch("scim2_client.engines.httpx.SyncSCIMClient.replace")
def test_failed_scim_group_update(
    scim_mock,
    testclient,
    scim_client_for_trusted_client,
    backend,
    caplog,
    bar_group,
):
    scim_mock.side_effect = mock.Mock(side_effect=Exception())

    backend.save(bar_group)

    assert (
        "canaille",
        logging.WARNING,
        "SCIM Group bar update for client Some client failed",
    ) in caplog.record_tuples


@mock.patch("scim2_client.engines.httpx.SyncSCIMClient.delete")
def test_failed_scim_group_delete(
    scim_mock,
    testclient,
    scim_client_for_trusted_client,
    backend,
    caplog,
    user,
):
    scim_mock.side_effect = mock.Mock(side_effect=Exception())

    group = models.Group(
        members=[user],
        display_name="foobar",
    )
    backend.save(group)

    backend.delete(group)

    assert (
        "canaille",
        logging.WARNING,
        "SCIM Group foobar delete for client Some client failed",
    ) in caplog.record_tuples


def test_user_from_canaille_to_scim_client_without_enterprise_user_extension(
    scim_client_for_trusted_client, user
):
    User = scim_client_for_trusted_client.get_resource_model("User")

    scim_user = user_from_canaille_to_scim_client(user, User, None)
    assert isinstance(scim_user, User)
    assert not isinstance(scim_user, MyUser[EnterpriseUser])
    assert EnterpriseUser not in scim_user



import datetime

from flask import current_app
from flask import g
from flask import session

from canaille.app import models


def current_user():
    if "user" in g:
        return g.user

    for user_id, login_datetime in session.get("user_id", [])[::-1]:
        user = current_app.backend.instance.get(models.User, user_id)
        if user and (
            not current_app.backend.has_account_lockability() or not user.locked
        ):
            g.user = user
            g.last_login_datetime = datetime.datetime.fromisoformat(login_datetime)
            return g.user

        session["user_id"].remove((user_id, login_datetime))

    if "user_id" in session and not session["user_id"]:
        del session["user_id"]

    return None


def current_user_login_datetime() -> datetime.datetime | None:
    if "last_login_datetime" in g:
        return g.last_login_datetime

    for user_id, login_datetime in session.get("user_id", [])[::-1]:
        user = current_app.backend.instance.get(models.User, user_id)
        if user and (
            not current_app.backend.has_account_lockability() or not user.locked
        ):
            g.user = user
            g.last_login_datetime = datetime.datetime.fromisoformat(login_datetime)
            return g.last_login_datetime

        session["user_id"].remove((user_id, login_datetime))

    if "user_id" in session and not session["user_id"]:
        del session["user_id"]

    return None


def login_user(user):
    """Open a session for the user."""
    g.user = user
    g.last_login_datetime = datetime.datetime.now(datetime.timezone.utc)
    try:
        previous = (
            session["user_id"]
            if isinstance(session["user_id"], list)
            else [session["user_id"]]
        )
        session["user_id"] = previous + [(user.id, g.last_login_datetime.isoformat())]
    except KeyError:
        session["user_id"] = [(user.id, g.last_login_datetime.isoformat())]


def logout_user():
    """Close the user session."""
    try:
        session["user_id"].pop()
        if not session["user_id"]:
            del session["user_id"]
    except (IndexError, KeyError):
        pass

    if "user" in g:
        del g.user

    if "last_login_datetime" in g:
        del g.last_login_datetime



from canaille.app.configuration import BaseModel


class SCIMSettings(BaseModel):
    """SCIM settings."""

    ENABLE_SERVER: bool = True
    """Whether the SCIM server API is enabled.

    When enabled, services plugged to Canaille can update users and groups using the API."""

    ENABLE_CLIENT: bool = False
    """Whether the state of :class:`~canaille.core.models.User` and :class:`~canaille.core.models.Group` are broadcasted to clients using the SCIM protocol.

    When enabled, any creation, edition or deletion of a client or a group will be replicated on clients that implement the SCIM protocol."""



import pytest

from canaille.app.configuration import settings_factory
from canaille.backends.ldap.backend import LDAPBackend
from tests.backends.ldap import CustomSlapdObject


@pytest.fixture(scope="session")
def slapd_server():
    slapd = CustomSlapdObject()

    try:
        slapd.start()
        slapd.init_tree()
        for ldif in (
            "dev/ldif/memberof-config.ldif",
            "dev/ldif/ppolicy-config.ldif",
            "dev/ldif/ppolicy.ldif",
            "dev/ldif/otp-config.ldif",
            "canaille/backends/ldap/schemas/oauth2-openldap.ldif",
            "dev/ldif/bootstrap-users-tree.ldif",
            "dev/ldif/bootstrap-oidc-tree.ldif",
        ):
            slapd.ldapadd(None, ["-f", ldif])
        yield slapd
    finally:
        slapd.stop()


@pytest.fixture
def ldap_configuration(configuration, slapd_server):
    configuration["CANAILLE"]["DATABASE"] = "ldap"
    configuration["CANAILLE_LDAP"] = {
        "ROOT_DN": slapd_server.suffix,
        "URI": slapd_server.ldap_uri,
        "BIND_DN": slapd_server.root_dn,
        "BIND_PW": slapd_server.root_pw,
        "USER_BASE": "ou=users",
        "USER_RDN": "uid",
        "GROUP_BASE": "ou=groups",
        "TIMEOUT": 0.1,
        "USER_CLASS": ["inetOrgPerson", "oathHOTPToken"],
    }
    yield configuration
    del configuration["CANAILLE_LDAP"]


@pytest.fixture
def ldap_backend(slapd_server, ldap_configuration):
    config_obj = settings_factory(ldap_configuration)
    config_dict = config_obj.model_dump()
    backend = LDAPBackend(config_dict)
    with backend.session():
        yield backend



from canaille.app import models


def test_model_references_set_unsaved_object(
    testclient, logged_moderator, user, backend
):
    """LDAP groups can be inconsistent by containing members which doesn't exist."""
    group = models.Group(members=[user], display_name="foo")
    backend.save(group)
    backend.reload(user)

    non_existent_user = models.User(
        formatted_name="foo", family_name="bar", user_name="baz"
    )
    group.members = group.members + [non_existent_user]
    assert group.members == [user, non_existent_user]

    backend.save(group)
    assert group.members == [user, non_existent_user]

    backend.reload(group)
    assert group.members == [user]

    testclient.get("/groups/foo", status=200)

    backend.delete(group)



import datetime
import logging
import re
from functools import wraps

from flask import abort
from flask import current_app
from flask import flash
from flask import make_response
from flask import redirect
from flask import request
from flask import session
from flask import url_for
from flask_caching import Cache
from flask_wtf.csrf import CSRFProtect
from werkzeug.exceptions import HTTPException
from werkzeug.routing import BaseConverter

from canaille.app.i18n import gettext as _
from canaille.app.session import current_user
from canaille.app.templating import render_template

csrf = CSRFProtect()
cache = Cache()


def user_needed(*args):
    permissions = set(args)

    def wrapper(view_function):
        @wraps(view_function)
        def decorator(*args, **kwargs):
            user = current_user()
            if not user or not user.can(*permissions):
                abort(403)

            if user.has_expired_password():
                flash(
                    _("Your password has expired, please choose a new password."),
                    "info",
                )
                return redirect(
                    url_for(
                        "core.account.reset",
                        user=user,
                    )
                )

            return view_function(*args, user=user, **kwargs)

        return decorator

    return wrapper


def smtp_needed():
    def wrapper(view_function):
        @wraps(view_function)
        def decorator(*args, **kwargs):
            if current_app.features.has_smtp:
                return view_function(*args, **kwargs)

            message = _("No SMTP server has been configured")
            logging.warning(message)
            return (
                render_template(
                    "error.html",
                    error_code=500,
                    icon="tools",
                    description=message,
                ),
                500,
            )

        return decorator

    return wrapper


def request_is_boosted() -> bool:
    """Whether the request is boosted with HTMX.

    https://htmx.org/docs/#boosting
    """
    return request.headers.get("HX-Boosted", False)


def request_is_partial() -> bool:
    """Whether the request only updates a subset of the DOM.

    True for form inline validation for instance.
    """
    return request.headers.get("HX-Request", False) and not request_is_boosted()


def partial_request_trigger() -> str | None:
    """Return the name of the form element that triggered the HTMX request."""
    base_name = request.headers.get("HX-Trigger-Name")
    return re.sub("-[0-9]+$", "", base_name) if base_name else base_name


def render_htmx_template(template, htmx_template=None, **kwargs):
    if request_is_partial():
        if htmx_template:
            template = htmx_template
        else:
            *dirs, file = template.split("/")
            template = "/".join([*dirs, "partial", file])
    return render_template(template, **kwargs)


def model_converter(model):
    class ModelConverter(BaseConverter):
        def __init__(self, *args, required=True, **kwargs):
            self.required = required
            super().__init__(self, *args, **kwargs)

        def to_url(self, instance):
            return instance.identifier or instance.id

        def to_python(self, identifier):
            current_app.backend.setup()
            instance = current_app.backend.get(model, identifier)
            if self.required and not instance:
                abort(404)

            return instance

    return ModelConverter


def redirect_to_bp_handlers(app, error):
    """Find and execute blueprint handlers matching an error.

    There is currently no way to make 404 handling generic:
    https://flask.palletsprojects.com/en/stable/errorhandling/#handling
        However, the blueprint cannot handle 404 routing errors because the
        404 occurs at the routing level before the blueprint can be determined.
    """
    for bp in app.blueprints.values():
        if bp.url_prefix and request.path.startswith(bp.url_prefix):
            for type_, handler in bp.error_handler_spec[None][None].items():
                if type_ in (error.code, HTTPException):  # pragma: no branch
                    return make_response(handler(error))
    return None


def setup_flask_blueprints(app):
    import canaille.core.endpoints

    app.url_map.strict_slashes = False

    app.register_blueprint(canaille.core.endpoints.bp)

    if app.features.has_oidc:  # pragma: no branch
        import canaille.oidc.endpoints

        app.register_blueprint(canaille.oidc.endpoints.bp)

    if app.features.has_scim_server:  # pragma: no branch
        import canaille.scim.endpoints

        app.register_blueprint(canaille.scim.endpoints.bp)


def setup_flask(app):
    from canaille.app.templating import render_template

    csrf.init_app(app)
    cache.init_app(app)

    @app.before_request
    def make_session_permanent():
        session.permanent = True
        app.permanent_session_lifetime = datetime.timedelta(days=365)

    @app.errorhandler(400)
    def bad_request(error):
        return render_template("error.html", description=error, error_code=400), 400

    @app.errorhandler(403)
    def unauthorized(error):
        return render_template("error.html", description=error, error_code=403), 403

    @app.errorhandler(404)
    def page_not_found(error):
        from canaille.app.flask import redirect_to_bp_handlers

        return redirect_to_bp_handlers(app, error) or render_template(
            "error.html", description=error, error_code=404
        ), 404

    @app.errorhandler(500)
    def server_error(error):  # pragma: no cover
        return render_template("error.html", description=error, error_code=500), 500


def setup_flask_converters(app):
    from canaille.app import models
    from canaille.app.flask import model_converter

    for model_name, model_class in models.MODELS.items():
        app.url_map.converters[model_name.lower()] = model_converter(model_class)



import datetime
from urllib.parse import parse_qs
from urllib.parse import urlsplit

from authlib.jose import jwt

from canaille.app import models

from . import client_credentials


def test_nominal_case(
    testclient, logged_user, client, keypair, trusted_client, backend, caplog
):
    assert not backend.query(models.Consent)

    res = testclient.get(
        "/oauth/authorize",
        params=dict(
            response_type="code",
            client_id=client.client_id,
            scope="openid profile email groups address phone",
            nonce="somenonce",
            tos_uri="https://client.test/tos",
            policy_uri="https://client.test/policy",
            redirect_uri="https://client.test/redirect1",
        ),
        status=200,
    )

    res = res.form.submit(name="answer", value="accept", status=302)

    assert res.location.startswith(client.redirect_uris[0])
    params = parse_qs(urlsplit(res.location).query)
    code = params["code"][0]

    res = testclient.post(
        "/oauth/token",
        params=dict(
            grant_type="authorization_code",
            code=code,
            scope="openid profile email groups address phone",
            redirect_uri=client.redirect_uris[0],
        ),
        headers={"Authorization": f"Basic {client_credentials(client)}"},
        status=200,
    )

    id_token = res.json["id_token"]
    claims = jwt.decode(id_token, keypair[1])
    assert claims.header["kid"]
    assert claims["sub"] == logged_user.user_name
    assert claims["name"] == logged_user.formatted_name
    assert claims["aud"] == [client.client_id, trusted_client.client_id]

    for consent in backend.query(models.Consent, client=client, subject=logged_user):
        backend.delete(consent)


def test_auth_time(
    testclient,
    user,
    client,
    keypair,
    backend,
):
    """Check that the ID token contains the user authentication time."""
    auth_time = datetime.datetime(2025, 1, 1, tzinfo=datetime.timezone.utc)
    with testclient.session_transaction() as session:
        session["user_id"] = [(user.id, auth_time.isoformat())]

    res = testclient.get(
        "/oauth/authorize",
        params=dict(
            response_type="code",
            client_id=client.client_id,
            scope="openid profile email groups address phone",
            nonce="somenonce",
            tos_uri="https://client.test/tos",
            policy_uri="https://client.test/policy",
            redirect_uri="https://client.test/redirect1",
            max_age=15000,
        ),
        status=200,
    )

    res = res.form.submit(name="answer", value="accept", status=302)
    params = parse_qs(urlsplit(res.location).query)
    code = params["code"][0]

    res = testclient.post(
        "/oauth/token",
        params=dict(
            grant_type="authorization_code",
            code=code,
            scope="openid profile email groups address phone",
            redirect_uri=client.redirect_uris[0],
        ),
        headers={"Authorization": f"Basic {client_credentials(client)}"},
        status=200,
    )

    id_token = res.json["id_token"]
    claims = jwt.decode(id_token, keypair[1])
    id_token_auth_time = datetime.datetime.fromtimestamp(
        claims["auth_time"], tz=datetime.timezone.utc
    )
    assert id_token_auth_time == auth_time

    backend.delete(backend.get(models.Consent))


def test_auth_time_update(
    testclient,
    user,
    client,
    keypair,
    backend,
):
    """Certification test 'oidcc-max-age-10000'.

    This test calls the authorization endpoint test twice. The first time it includes max_age=15000 (so that the OP is required to return auth_time in the id_token). The second time it includes max_age=10000, and the authorization server must not request that the user logs in. The test verifies that auth_time and sub are consistent between the id_tokens from the first and second authorizations.
    """
    auth_time = datetime.datetime(2025, 1, 1, tzinfo=datetime.timezone.utc)
    with testclient.session_transaction() as session:
        session["user_id"] = [(user.id, auth_time.isoformat())]

    res = testclient.get(
        "/oauth/authorize",
        params=dict(
            response_type="code",
            client_id=client.client_id,
            scope="openid profile email groups address phone",
            nonce="somenonce",
            tos_uri="https://client.test/tos",
            policy_uri="https://client.test/policy",
            redirect_uri="https://client.test/redirect1",
            max_age=15000,
        ),
        status=200,
    )

    res = res.form.submit(name="answer", value="accept", status=302)
    params = parse_qs(urlsplit(res.location).query)
    code1 = params["code"][0]

    res = testclient.post(
        "/oauth/token",
        params=dict(
            grant_type="authorization_code",
            code=code1,
            scope="openid profile email groups address phone",
            redirect_uri=client.redirect_uris[0],
        ),
        headers={"Authorization": f"Basic {client_credentials(client)}"},
        status=200,
    )

    id_token = res.json["id_token"]
    claims = jwt.decode(id_token, keypair[1])
    id_token_1_auth_time = datetime.datetime.fromtimestamp(
        claims["auth_time"], tz=datetime.timezone.utc
    )
    assert id_token_1_auth_time == auth_time

    backend.delete(backend.get(models.Consent))

    res = testclient.get(
        "/oauth/authorize",
        params=dict(
            response_type="code",
            client_id=client.client_id,
            scope="openid profile email groups address phone",
            nonce="somenonce",
            tos_uri="https://client.test/tos",
            policy_uri="https://client.test/policy",
            redirect_uri="https://client.test/redirect1",
        ),
        status=200,
    )

    res = res.form.submit(name="answer", value="accept", status=302)
    params = parse_qs(urlsplit(res.location).query)
    code2 = params["code"][0]

    assert code1 != code2

    res = testclient.post(
        "/oauth/token",
        params=dict(
            grant_type="authorization_code",
            code=code2,
            scope="openid profile email groups address phone",
            redirect_uri=client.redirect_uris[0],
            max_age=10000,
        ),
        headers={"Authorization": f"Basic {client_credentials(client)}"},
        status=200,
    )

    id_token = res.json["id_token"]
    claims = jwt.decode(id_token, keypair[1])
    id_token_2_auth_time = datetime.datetime.fromtimestamp(
        claims["auth_time"], tz=datetime.timezone.utc
    )
    assert id_token_2_auth_time == auth_time



import uuid

from canaille import create_app  # pragma: no cover

app = create_app(env_file=".env")  # pragma: no cover

if app.config["SECRET_KEY"] is None:  # pragma: no cover
    app.logger.warning("Missing 'SECRET_KEY' configuration parameter.")
    app.config["SECRET_KEY"] = str(uuid.uuid4())



import itertools
from collections.abc import Iterable


class LDAPObjectQuery:
    def __init__(self, klass, items):
        self.klass = klass
        self.items = items

    def __len__(self):
        return len(self.items)

    def __getitem__(self, item):
        if isinstance(item, slice):
            return (self.decorate(obj[1]) for obj in self.items[item])

        return self.decorate(self.items[item][1])

    def __iter__(self):
        return (self.decorate(obj[1]) for obj in self.items)

    def __eq__(self, other):
        if isinstance(other, Iterable):
            return all(
                a == b
                for a, b in itertools.zip_longest(
                    iter(self), iter(other), fillvalue=object()
                )
            )
        return super().__eq__(other)

    def __bool__(self):
        return bool(self.items)

    def decorate(self, args):
        klass = self.guess_class(self.klass, args["objectClass"])
        obj = klass()
        obj.state = args
        obj.exists = True
        return obj

    def guess_class(self, klass, object_classes):
        from .ldapobject import LDAPObject
        from .ldapobject import LDAPObjectMetaclass

        if klass == LDAPObject:
            models = [
                LDAPObjectMetaclass.ldap_to_python_class[oc.decode()]
                for oc in object_classes
                if oc.decode() in LDAPObjectMetaclass.ldap_to_python_class
            ]
            return models[0]
        return klass



import logging
import os
import pathlib
from unittest import mock

import pytest
import tomlkit
from flask_webtest import TestApp
from pydantic import ValidationError

from canaille import create_app
from canaille.app.configuration import CheckResult
from canaille.app.configuration import check_network_config
from canaille.app.configuration import check_smpp_connection
from canaille.app.configuration import check_smtp_connection
from canaille.app.configuration import export_config
from canaille.app.configuration import sanitize_rst_text
from canaille.app.configuration import settings_factory


def test_configuration_secrets_directory(tmp_path, backend, configuration):
    os.environ["SECRETS_DIR"] = str(tmp_path)

    secret_key_path = tmp_path / "SECRET_KEY"
    with open(secret_key_path, "w") as fd:
        fd.write("very-secret")

    del configuration["SECRET_KEY"]

    app = create_app(configuration)
    assert app.config["SECRET_KEY"] == "very-secret"
    del os.environ["SECRETS_DIR"]


@pytest.mark.skip
# Not fully implemented in pydantic-settings yet
# https://github.com/pydantic/pydantic-settings/issues/154
def test_configuration_nestedsecrets_directory(tmp_path, backend, configuration):
    os.environ["SECRETS_DIR"] = str(tmp_path)

    smtp_password_path = tmp_path / "CANAILLE__SMTP__PASSWORD"
    with open(smtp_password_path, "w") as fd:
        fd.write("very-very-secret")

    del configuration["CANAILLE"]["SMTP"]["PASSWORD"]

    app = create_app(configuration)
    assert app.config["CANAILLE"]["SMTP"]["PASSWORD"] == "very-very-secret"
    del os.environ["SECRETS_DIR"]


def test_no_configuration(configuration, tmp_path):
    """Test that no configuration still makes valid Canaille application."""
    os.environ["DEBUG"] = "1"

    app = create_app()
    assert app.config["CANAILLE"]["NAME"] == "Canaille"

    del os.environ["DEBUG"]


def test_environment_configuration(configuration, tmp_path):
    """Test loading the configuration from a toml file passed by the CONFIG environment var."""
    config_path = os.path.join(tmp_path, "config.toml")
    with open(config_path, "w") as fd:
        tomlkit.dump(configuration, fd)

    os.environ["CONFIG"] = config_path
    app = create_app()
    assert app.config["CANAILLE"]["SMTP"]["FROM_ADDR"] == "admin@mydomain.test"

    del os.environ["CONFIG"]
    os.remove(config_path)


def test_local_configuration(configuration, tmp_path):
    """Test loading the configuration from a local config.toml file."""
    cwd = os.getcwd()
    os.chdir(tmp_path)

    config_path = os.path.join(tmp_path, "canaille.toml")
    with open(config_path, "w") as fd:
        tomlkit.dump(configuration, fd)

    app = create_app()
    assert app.config["CANAILLE"]["SMTP"]["FROM_ADDR"] == "admin@mydomain.test"

    os.chdir(cwd)
    os.remove(config_path)


def test_configuration_from_environment_vars(tmp_path):
    """Canaille should read configuration from environment vars."""
    os.environ["SECRET_KEY"] = "very-very-secret"
    os.environ["CANAILLE__SMTP__FROM_ADDR"] = "user@mydomain.test"
    os.environ["CANAILLE_SQL__DATABASE_URI"] = f"sqlite:///{tmp_path}/anything.db"

    conf = settings_factory({"TIMEZONE": "UTC"})
    assert conf.SECRET_KEY == "very-very-secret"
    assert conf.CANAILLE.SMTP.FROM_ADDR == "user@mydomain.test"
    assert conf.CANAILLE_SQL.DATABASE_URI == f"sqlite:///{tmp_path}/anything.db"

    app = create_app({"TIMEZONE": "UTC"})
    assert app.config["SECRET_KEY"] == "very-very-secret"
    assert app.config["CANAILLE"]["SMTP"]["FROM_ADDR"] == "user@mydomain.test"
    assert (
        app.config["CANAILLE_SQL"]["DATABASE_URI"]
        == f"sqlite:///{tmp_path}/anything.db"
    )

    del os.environ["SECRET_KEY"]
    del os.environ["CANAILLE__SMTP__FROM_ADDR"]
    del os.environ["CANAILLE_SQL__DATABASE_URI"]


def test_disable_env_var_loading(tmp_path, configuration):
    """Canaille should not read configuration from environment vars when env_prefix is False."""
    del configuration["SERVER_NAME"]
    os.environ["SERVER_NAME"] = "example.test"
    os.environ["FOOBAR_SERVER_NAME"] = "foobar.example.test"

    app = create_app(configuration, env_prefix="")
    assert app.config["SERVER_NAME"] == "example.test"

    app = create_app(configuration, env_prefix="FOOBAR_")
    assert app.config["SERVER_NAME"] == "foobar.example.test"

    del os.environ["SERVER_NAME"]


def test_dotenv_file(tmp_path, configuration):
    """Canaille should read configuration from .env files."""
    os.environ["ENV_FILE"] = ".env"
    oldcwd = os.getcwd()
    os.chdir(tmp_path)
    dotenv = tmp_path / ".env"

    with open(dotenv, "w") as fd:
        fd.write("FOOBAR=custom-value")

    app = create_app(configuration)
    assert app.config["FOOBAR"] == "custom-value"
    os.chdir(oldcwd)
    del os.environ["ENV_FILE"]


def test_custom_dotenv_file(tmp_path, configuration):
    """Canaille should read configuration from custom .env files if they are passed with env_file."""
    dotenv = tmp_path / "custom.env"
    with open(dotenv, "w") as fd:
        fd.write("FOOBAR=other-custom-value")

    app = create_app(configuration, env_file=dotenv)
    assert app.config["FOOBAR"] == "other-custom-value"


def test_disable_dotenv_file(tmp_path, configuration):
    """Canaille should ignore .env files if env_file is None."""
    oldcwd = os.getcwd()
    os.chdir(tmp_path)
    dotenv = tmp_path / ".env"

    with open(dotenv, "w") as fd:
        fd.write("FOOBAR=custom-value")

    app = create_app(configuration, env_file=None)
    assert "FOOBAR" not in app.config
    os.chdir(oldcwd)


def test_smtp_connection_remote_smtp_unreachable(testclient, backend, configuration):
    configuration["CANAILLE"]["SMTP"]["HOST"] = "smtp://invalid-smtp.com"
    config_obj = settings_factory(configuration)
    config_dict = config_obj.model_dump()["CANAILLE"]["SMTP"]
    assert check_smtp_connection(config_dict) == CheckResult(
        success=False,
        message=f"Could not connect to the SMTP server 'smtp://invalid-smtp.com' on port '{configuration['CANAILLE']['SMTP']['PORT']}'",
    )


def test_smtp_connection_remote_smtp_wrong_credentials(
    testclient, backend, configuration
):
    configuration["CANAILLE"]["SMTP"]["PASSWORD"] = "invalid-password"
    config_obj = settings_factory(configuration)
    config_dict = config_obj.model_dump()["CANAILLE"]["SMTP"]
    assert check_smtp_connection(config_dict) == CheckResult(
        success=False, message="SMTP authentication failed with user 'user'"
    )


def test_smtp_connection_remote_smtp_no_credentials(
    testclient, backend, configuration, mock_smpp
):
    del configuration["CANAILLE"]["SMTP"]["LOGIN"]
    del configuration["CANAILLE"]["SMTP"]["PASSWORD"]
    config_obj = settings_factory(configuration)
    config_dict = config_obj.model_dump()["CANAILLE"]["SMTP"]
    assert check_smtp_connection(config_dict) == CheckResult(
        success=True, message="Successful SMTP connection"
    )


def test_smtp_bad_tls(testclient, backend, smtpd, configuration):
    configuration["CANAILLE"]["SMTP"]["TLS"] = False
    config_obj = settings_factory(configuration)
    config_dict = config_obj.model_dump()["CANAILLE"]["SMTP"]
    assert check_smtp_connection(config_dict) == CheckResult(
        success=False, message="SMTP AUTH extension not supported by server."
    )


def test_smtp_discovery_success(testclient, backend, configuration, smtpd):
    """If a SMTP server is configured on the default port and is accessible without authentication, it should be used."""
    del configuration["CANAILLE"]["SMTP"]
    with mock.patch("canaille.core.configuration.DEFAULT_SMTP_PORT", smtpd.port):
        config_obj = settings_factory(configuration)
        config_dict = config_obj.model_dump()["CANAILLE"]["SMTP"]
        assert check_smtp_connection(config_dict) == CheckResult(
            success=True, message="Successful SMTP connection"
        )


def test_smtp_discovery_failure(testclient, backend, configuration, smtpd):
    """Test that something goes wrong during the SMTP server discovery."""
    del configuration["CANAILLE"]["SMTP"]
    with mock.patch("canaille.core.configuration.DEFAULT_SMTP_PORT", 99999):
        config_obj = settings_factory(configuration)
        assert config_obj.model_dump()["CANAILLE"]["SMTP"] is None


@pytest.fixture
def themed_testclient(app, configuration, backend):
    root = os.path.dirname(os.path.abspath(__file__))
    test_theme_path = os.path.join(root, "fixtures", "themes", "test")
    configuration["CANAILLE"]["THEME"] = test_theme_path

    app = create_app(configuration)

    return TestApp(app)


def test_theme(testclient, themed_testclient, backend):
    res = testclient.get("/login")
    res.mustcontain(no="TEST_THEME")

    res = themed_testclient.get("/login")
    res.mustcontain("TEST_THEME")


def test_invalid_theme(configuration, backend):
    with pytest.raises(
        ValidationError,
        match=r"Path does not point to a directory",
    ):
        configuration["CANAILLE"]["THEME"] = "invalid"
        settings_factory(configuration)

    with pytest.raises(
        ValidationError,
        match=r"Path does not point to a directory",
    ):
        configuration["CANAILLE"]["THEME"] = "/path/to/invalid"
        settings_factory(configuration)


def test_enable_password_compromission_check_with_and_without_admin_email(
    configuration, backend
):
    configuration["CANAILLE"]["ENABLE_PASSWORD_COMPROMISSION_CHECK"] = False
    configuration["CANAILLE"]["ADMIN_EMAIL"] = None
    settings_factory(configuration)

    configuration["CANAILLE"]["ENABLE_PASSWORD_COMPROMISSION_CHECK"] = True
    configuration["CANAILLE"]["ADMIN_EMAIL"] = "admin_default_mail@mydomain.test"
    settings_factory(configuration)

    with pytest.raises(
        ValidationError,
        match=r"You must set an administration email if you want to check if users' passwords are compromised.",
    ):
        configuration["CANAILLE"]["ENABLE_PASSWORD_COMPROMISSION_CHECK"] = True
        configuration["CANAILLE"]["ADMIN_EMAIL"] = None
        settings_factory(configuration)


def test_invalid_otp_option(configuration, backend):
    with pytest.raises(
        ValidationError,
        match=r"Input should be 'TOTP' or 'HOTP'",
    ):
        configuration["CANAILLE"]["OTP_METHOD"] = "invalid"
        settings_factory(configuration)


def test_email_otp_without_smtp(configuration, backend):
    with pytest.raises(
        ValidationError,
        match=r"Cannot activate email one-time password authentication without SMTP",
    ):
        configuration["CANAILLE"]["SMTP"] = None
        configuration["CANAILLE"]["EMAIL_OTP"] = True
        settings_factory(configuration)


def test_sms_otp_without_smpp(configuration, backend):
    with pytest.raises(
        ValidationError,
        match=r"Cannot activate sms one-time password authentication without SMPP",
    ):
        configuration["CANAILLE"]["SMPP"] = None
        configuration["CANAILLE"]["SMS_OTP"] = True
        settings_factory(configuration)


def test_smpp_connection_remote_smpp_unreachable(testclient, backend, configuration):
    configuration["CANAILLE"]["SMPP"] = {
        "HOST": "invalid-smpp.com",
        "PORT": 2775,
        "LOGIN": "user",
        "PASSWORD": "user",
    }
    config_obj = settings_factory(configuration)
    config_dict = config_obj.model_dump()["CANAILLE"]["SMPP"]
    assert check_smpp_connection(config_dict) == CheckResult(
        success=False,
        message="Could not connect to the SMPP server 'invalid-smpp.com' on port '2775'",
    )


def test_check_network_config_without_smpp(configuration, backend, mock_smpp):
    configuration["CANAILLE"]["SMPP"] = None
    config_obj = settings_factory(configuration)
    config_dict = config_obj.model_dump()

    assert CheckResult(
        success=None, message="No SMPP server configured"
    ) in check_network_config(config_dict)


def test_smpp_connection_remote_smpp_no_credentials(
    testclient, backend, configuration, mock_smpp
):
    del configuration["CANAILLE"]["SMPP"]["LOGIN"]
    del configuration["CANAILLE"]["SMPP"]["PASSWORD"]
    config_obj = settings_factory(configuration)
    config_dict = config_obj.model_dump()["CANAILLE"]["SMPP"]
    assert check_smpp_connection(config_dict) == CheckResult(
        success=True, message="Successful SMPP connection"
    )


def test_no_secret_key(configuration, caplog):
    del configuration["SECRET_KEY"]

    os.environ["DEBUG"] = "1"
    from canaille.app.server import app

    assert (
        "canaille",
        logging.WARNING,
        "Missing 'SECRET_KEY' configuration parameter.",
    ) in caplog.record_tuples

    testclient = TestApp(app)
    res = testclient.get("/login")
    res.mustcontain(
        "Your Canaille instance is not fully configured and not ready for production."
    )
    del os.environ["DEBUG"]


def test_sanitize_rst():
    assert sanitize_rst_text("``somevar``") == "somevar"
    assert sanitize_rst_text(":class:`~canaille.core.models.User`") == "User"
    assert sanitize_rst_text(":data:`None`") == "None"
    assert sanitize_rst_text(":py:data:`None`") == "None"
    assert (
        sanitize_rst_text("`Yaal Coop <https://yaal.coop>`_")
        == "Yaal Coop (https://yaal.coop)"
    )

    assert (
        sanitize_rst_text(""".. code-block:: python

    var = "value"
""")
        == '    var = "value"\n'
    )
    assert (
        sanitize_rst_text(""".. danger::

    foobar
""")
        == "    foobar\n"
    )


def test_export_current_config(backend, configuration, tmp_path):
    """Check the configuration TOML export with the current app configuration."""
    if "memory" not in backend.__class__.__module__:
        pytest.skip()

    toml_export = tmp_path / "config.toml"

    configuration["SECRET_KEY"] = "very-secret"
    configuration["CANAILLE"]["SMTP"]["PORT"] = 25

    config_obj = settings_factory(configuration, init_with_examples=True)
    export_config(config_obj, toml_export)

    toml_expected = (
        pathlib.Path(__file__).parent / "fixtures" / "current-app-config.toml"
    )

    with open(toml_export) as fd:
        actual_content = fd.read()

    with open(toml_expected) as fd:
        expected_content = fd.read()

    assert actual_content == expected_content


def test_export_default_config(tmp_path, backend):
    """Check the configuration TOML export with the default configuration."""
    toml_export = tmp_path / "config.toml"

    config_obj = settings_factory(init_with_examples=True)
    export_config(config_obj, toml_export)

    toml_expected = pathlib.Path(__file__).parent / "fixtures" / "default-config.toml"

    with open(toml_export) as fd:
        actual_content = fd.read()

    with open(toml_expected) as fd:
        expected_content = fd.read()

    assert actual_content == expected_content



import os

import slapd


class CustomSlapdObject(slapd.Slapd):
    def __init__(self):
        schemas = [
            schema
            for schema in [
                "core.ldif",
                "cosine.ldif",
                "nis.ldif",
                "inetorgperson.ldif",
                "ppolicy.ldif",
            ]
            if os.path.exists(os.path.join(self.SCHEMADIR, schema))
        ]

        super().__init__(
            suffix="dc=example,dc=org",
            schemas=schemas,
        )



import json
from datetime import datetime
from unittest import mock

from authlib.jose import jwt

from canaille.app import models


def test_client_registration_with_authentication_static_token(
    testclient, backend, client, user
):
    assert not testclient.app.config["CANAILLE_OIDC"].get(
        "DYNAMIC_CLIENT_REGISTRATION_OPEN"
    )
    testclient.app.config["CANAILLE_OIDC"]["DYNAMIC_CLIENT_REGISTRATION_TOKENS"] = [
        "static-token"
    ]

    payload = {
        "redirect_uris": [
            "https://client.test/callback",
            "https://client.test/callback2",
        ],
        "post_logout_redirect_uris": [
            "https://client.test/logout_callback",
        ],
        "client_name": "My Example Client",
        "token_endpoint_auth_method": "client_secret_basic",
        "logo_uri": "https://client.test/logo.webp",
        "jwks_uri": "https://client.test/my_public_keys.jwks",
        "grant_types": ["authorization_code"],
        "response_types": ["code"],
    }
    headers = {"Authorization": "Bearer static-token"}

    res = testclient.post_json("/oauth/register", payload, headers=headers, status=201)
    client = backend.get(models.Client, client_id=res.json["client_id"])

    assert res.json == {
        "client_id": client.client_id,
        "client_secret": client.client_secret,
        "client_id_issued_at": mock.ANY,
        "client_name": "My Example Client",
        "client_secret_expires_at": 0,
        "jwks_uri": "https://client.test/my_public_keys.jwks",
        "logo_uri": "https://client.test/logo.webp",
        "redirect_uris": [
            "https://client.test/callback",
            "https://client.test/callback2",
        ],
        "token_endpoint_auth_method": "client_secret_basic",
        "grant_types": ["authorization_code"],
        "response_types": ["code"],
        "scope": "openid",
        "application_type": "web",
        "id_token_signed_response_alg": "RS256",
        "require_auth_time": False,
    }

    assert client.client_name == "My Example Client"
    assert client.redirect_uris == [
        "https://client.test/callback",
        "https://client.test/callback2",
    ]
    assert client.post_logout_redirect_uris == [
        "https://client.test/logout_callback",
    ]
    assert client.token_endpoint_auth_method == "client_secret_basic"
    assert client.logo_uri == "https://client.test/logo.webp"
    assert client.jwks_uri == "https://client.test/my_public_keys.jwks"
    assert client in client.audience
    backend.delete(client)


def test_client_registration_with_uri_fragments(testclient, backend, client, user):
    testclient.app.config["CANAILLE_OIDC"]["DYNAMIC_CLIENT_REGISTRATION_TOKENS"] = [
        "static-token"
    ]

    payload = {
        "redirect_uris": [
            "https://client.example.test/callback",
            "https://client.example.test/callback2#foo",
        ],
        "post_logout_redirect_uris": [
            "https://client.example.test/logout_callback",
        ],
        "client_name": "My Example Client",
        "token_endpoint_auth_method": "client_secret_basic",
        "logo_uri": "https://client.example.test/logo.webp",
        "jwks_uri": "https://client.example.test/my_public_keys.jwks",
        "grant_types": ["authorization_code"],
        "response_types": ["code"],
    }
    headers = {"Authorization": "Bearer static-token"}

    res = testclient.post_json("/oauth/register", payload, headers=headers, status=400)

    assert res.json == {
        "error_description": "Invalid claim 'redirect_uris'",
        "error": "invalid_client_metadata",
    }


def test_client_registration_with_authentication_no_token(
    testclient, backend, client, user
):
    assert not testclient.app.config["CANAILLE_OIDC"].get(
        "DYNAMIC_CLIENT_REGISTRATION_OPEN"
    )

    payload = {
        "redirect_uris": [
            "https://client.test/callback",
            "https://client.test/callback2",
        ],
        "client_name": "My Example Client",
        "token_endpoint_auth_method": "client_secret_basic",
        "logo_uri": "https://client.test/logo.webp",
        "jwks_uri": "https://client.test/my_public_keys.jwks",
        "grant_types": ["authorization_code"],
        "response_types": ["code"],
    }

    res = testclient.post_json("/oauth/register", payload, status=400)
    assert res.json == {
        "error": "access_denied",
        "error_description": "The resource owner or authorization server denied the request",
    }

    headers = {"Authorization": "Bearer"}
    res = testclient.post_json("/oauth/register", payload, headers=headers, status=400)
    assert res.json == {
        "error": "access_denied",
        "error_description": "The resource owner or authorization server denied the request",
    }


def test_client_registration_with_authentication_invalid_token(
    testclient, backend, client, user
):
    assert not testclient.app.config["CANAILLE_OIDC"].get(
        "DYNAMIC_CLIENT_REGISTRATION_OPEN"
    )

    payload = {
        "redirect_uris": [
            "https://client.test/callback",
            "https://client.test/callback2",
        ],
        "client_name": "My Example Client",
        "token_endpoint_auth_method": "client_secret_basic",
        "logo_uri": "https://client.test/logo.webp",
        "jwks_uri": "https://client.test/my_public_keys.jwks",
        "grant_types": ["authorization_code"],
        "response_types": ["code"],
    }
    headers = {"Authorization": "Bearer invalid-token"}
    res = testclient.post_json("/oauth/register", payload, headers=headers, status=400)
    assert res.json == {
        "error": "access_denied",
        "error_description": "The resource owner or authorization server denied the request",
    }


def test_client_registration_with_software_statement(testclient, backend, keypair):
    private_key, _ = keypair
    testclient.app.config["CANAILLE_OIDC"]["DYNAMIC_CLIENT_REGISTRATION_OPEN"] = True

    software_statement_payload = {
        "software_id": "4NRB1-0XZABZI9E6-5SM3R",
        "client_name": "Example Statement-based Client",
        "client_uri": "https://client.test/",
        "response_types": ["code"],
        "grant_types": ["authorization_code"],
    }
    software_statement_header = {"alg": "RS256"}
    software_statement = jwt.encode(
        software_statement_header, software_statement_payload, private_key
    ).decode()

    def test_client_registration_with_software_statement_with_different_scopes(
        scope_value, **kwargs
    ):
        payload = kwargs

        res = testclient.post_json("/oauth/register", payload, status=201)

        client = backend.get(models.Client, client_id=res.json["client_id"])
        assert res.json == {
            "client_id": client.client_id,
            "client_secret": client.client_secret,
            "client_id_issued_at": mock.ANY,
            "client_secret_expires_at": 0,
            "redirect_uris": [
                "https://client.test/callback",
                "https://client.test/callback2",
            ],
            "grant_types": ["authorization_code"],
            "response_types": ["code"],
            "scope": scope_value,
            "token_endpoint_auth_method": "client_secret_basic",
            "client_name": "Example Statement-based Client",
            "client_uri": "https://client.test/",
            "software_id": "4NRB1-0XZABZI9E6-5SM3R",
            "application_type": "web",
            "id_token_signed_response_alg": "RS256",
            "require_auth_time": False,
        }
        assert client.redirect_uris == [
            "https://client.test/callback",
            "https://client.test/callback2",
        ]
        assert client.token_endpoint_auth_method == "client_secret_basic"
        backend.delete(client)

    test_client_registration_with_software_statement_with_different_scopes(
        "openid profile",
        redirect_uris=[
            "https://client.test/callback",
            "https://client.test/callback2",
        ],
        software_statement=software_statement,
        scope="openid profile",
    )
    test_client_registration_with_software_statement_with_different_scopes(
        "openid",
        redirect_uris=[
            "https://client.test/callback",
            "https://client.test/callback2",
        ],
        software_statement=software_statement,
    )


def test_client_registration_without_authentication_ok(testclient, backend):
    testclient.app.config["CANAILLE_OIDC"]["DYNAMIC_CLIENT_REGISTRATION_OPEN"] = True

    payload = {
        "redirect_uris": [
            "https://client.test/callback",
            "https://client.test/callback2",
        ],
        "client_name": "My Example Client",
        "client_uri": "https://client.test",
        "token_endpoint_auth_method": "client_secret_basic",
        "logo_uri": "https://client.test/logo.webp",
        "jwks_uri": "https://client.test/my_public_keys.jwks",
        "grant_types": ["authorization_code", "implicit"],
        "response_types": ["code", "token"],
        "scope": "openid profile",
        "contacts": ["contact@test"],
        "tos_uri": "https://client.test/uri",
        "policy_uri": "https://client.test/policy",
        "software_id": "example",
        "software_version": "x.y.z",
    }

    res = testclient.post_json("/oauth/register", payload, status=201)

    client = backend.get(models.Client, client_id=res.json["client_id"])
    assert res.json == {
        "client_id": mock.ANY,
        "client_secret": mock.ANY,
        "client_id_issued_at": mock.ANY,
        "client_name": "My Example Client",
        "client_uri": "https://client.test",
        "client_secret_expires_at": 0,
        "jwks_uri": "https://client.test/my_public_keys.jwks",
        "logo_uri": "https://client.test/logo.webp",
        "redirect_uris": [
            "https://client.test/callback",
            "https://client.test/callback2",
        ],
        "token_endpoint_auth_method": "client_secret_basic",
        "grant_types": ["authorization_code", "implicit"],
        "response_types": ["code", "token"],
        "scope": "openid profile",
        "contacts": ["contact@test"],
        "tos_uri": "https://client.test/uri",
        "policy_uri": "https://client.test/policy",
        "software_id": "example",
        "software_version": "x.y.z",
        "application_type": "web",
        "id_token_signed_response_alg": "RS256",
        "require_auth_time": False,
    }
    assert client.client_name == "My Example Client"
    assert client.client_uri == "https://client.test"
    assert client.redirect_uris == [
        "https://client.test/callback",
        "https://client.test/callback2",
    ]
    assert client.token_endpoint_auth_method == "client_secret_basic"
    assert client.logo_uri == "https://client.test/logo.webp"
    assert client.jwks_uri == "https://client.test/my_public_keys.jwks"
    assert client.grant_types == ["authorization_code", "implicit"]
    assert client.response_types == ["code", "token"]
    assert client.scope == ["openid", "profile"]
    assert client.contacts == ["contact@test"]
    assert client.tos_uri == "https://client.test/uri"
    assert client.policy_uri == "https://client.test/policy"
    assert client.software_id == "example"
    assert client.software_version == "x.y.z"
    backend.delete(client)


def test_client_registration_with_jwks(testclient, backend):
    """Nominal test case for a registration with the 'jwks' parameter."""
    testclient.app.config["CANAILLE_OIDC"]["DYNAMIC_CLIENT_REGISTRATION_OPEN"] = True

    payload = {
        "redirect_uris": [
            "https://client.test/callback",
            "https://client.test/callback2",
        ],
        "client_name": "My Example Client",
        "client_uri": "https://client.test",
        "token_endpoint_auth_method": "client_secret_basic",
        "logo_uri": "https://client.test/logo.webp",
        "grant_types": ["authorization_code", "implicit"],
        "response_types": ["code", "token"],
        "scope": "openid profile",
        "jwks": {
            "keys": [
                {
                    "alg": "RS256",
                    "e": "AQAB",
                    "kty": "RSA",
                    "n": "wbLxLf5qi3iO_3pQPbulxfPm7p5Ameeow-On-ssQBjkaOrK9ZLHQZtCDxzEw VGmWPIe5jRx3Ot97PPHZz2ldvKN6rLlG7YiXCiijuz_an-ppWC52xa0Ue6l5iSIxS7Ot6WWyxgP0wA3JrDy85TNUYZ1O3hWSSJJjgO9RpY2JZLW_UVQFOy9HdsUHio46eTQ_vCqP9sK gRz3W5Al82ZL1iZhKye86FbgHIG4SXGjQB0kopT6DEjz_Bf-rxGmD9mu7Fx6DSn7qEXQZja35ELAvtuasYHvpMYCFUXLIlzN8H_HmsMfN-Fai7_FFsuss6Cpqt0NJUSrqxRZGOsoLj4 icJw",
                    "use": "sig",
                }
            ]
        },
    }
    res = testclient.post_json("/oauth/register", payload, status=201)

    client = backend.get(models.Client, client_id=res.json["client_id"])
    assert res.json["jwks"] == {
        "keys": [
            {
                "alg": "RS256",
                "e": "AQAB",
                "kty": "RSA",
                "n": "wbLxLf5qi3iO_3pQPbulxfPm7p5Ameeow-On-ssQBjkaOrK9ZLHQZtCDxzEw VGmWPIe5jRx3Ot97PPHZz2ldvKN6rLlG7YiXCiijuz_an-ppWC52xa0Ue6l5iSIxS7Ot6WWyxgP0wA3JrDy85TNUYZ1O3hWSSJJjgO9RpY2JZLW_UVQFOy9HdsUHio46eTQ_vCqP9sK gRz3W5Al82ZL1iZhKye86FbgHIG4SXGjQB0kopT6DEjz_Bf-rxGmD9mu7Fx6DSn7qEXQZja35ELAvtuasYHvpMYCFUXLIlzN8H_HmsMfN-Fai7_FFsuss6Cpqt0NJUSrqxRZGOsoLj4 icJw",
                "use": "sig",
            }
        ]
    }
    assert json.loads(client.jwks) == {
        "keys": [
            {
                "alg": "RS256",
                "e": "AQAB",
                "kty": "RSA",
                "n": "wbLxLf5qi3iO_3pQPbulxfPm7p5Ameeow-On-ssQBjkaOrK9ZLHQZtCDxzEw VGmWPIe5jRx3Ot97PPHZz2ldvKN6rLlG7YiXCiijuz_an-ppWC52xa0Ue6l5iSIxS7Ot6WWyxgP0wA3JrDy85TNUYZ1O3hWSSJJjgO9RpY2JZLW_UVQFOy9HdsUHio46eTQ_vCqP9sK gRz3W5Al82ZL1iZhKye86FbgHIG4SXGjQB0kopT6DEjz_Bf-rxGmD9mu7Fx6DSn7qEXQZja35ELAvtuasYHvpMYCFUXLIlzN8H_HmsMfN-Fai7_FFsuss6Cpqt0NJUSrqxRZGOsoLj4 icJw",
                "use": "sig",
            }
        ]
    }

    backend.delete(client)


def test_client_registration_with_all_attributes(testclient, backend, user):
    assert not testclient.app.config["CANAILLE_OIDC"].get(
        "DYNAMIC_CLIENT_REGISTRATION_OPEN"
    )
    testclient.app.config["CANAILLE_OIDC"]["DYNAMIC_CLIENT_REGISTRATION_TOKENS"] = [
        "static-token"
    ]

    payload = {
        "redirect_uris": [
            "https://client.test/callback",
            "https://client.test/callback2",
        ],
        "client_name": "My Example Client",
        "client_uri": "https://client.test",
        "token_endpoint_auth_method": "client_secret_basic",
        "token_endpoint_auth_signing_alg": "RS256",
        "logo_uri": "https://client.test/logo.webp",
        "jwks_uri": "https://client.test/my_public_keys.jwks",
        "grant_types": ["authorization_code"],
        "response_types": ["code"],
        "tos_uri": "https://client.test/uri",
        "policy_uri": "https://client.test/policy",
        "sector_identifier_uri": "https://client.test/sector.json",
        "subject_type": "public",
        "software_id": "example-software",
        "software_version": "x.y.z",
        "application_type": "web",
        "id_token_signed_response_alg": "RS256",
        "id_token_encrypted_response_alg": "RS256",
        "id_token_encrypted_response_enc": "A128CBC-HS256",
        "userinfo_signed_response_alg": "RS256",
        "userinfo_encrypted_response_alg": "RS256",
        "userinfo_encrypted_response_enc": "A128CBC-HS256",
        "default_max_age": 12345,
        "require_auth_time": True,
        "default_acr_values": ["basic", "mfa", "high"],
        "initiate_login_uri": "https://client.test/login",
        "request_object_signing_alg": "RS256",
        "request_object_encryption_alg": "RS256",
        "request_object_encryption_enc": "A128CBC-HS256",
        "request_uris": [
            "https://client.test/request_objects_1",
            "https://client.test/request_objects_2",
        ],
        "client_secret_expires_at": 0,
        "scope": "openid",
    }
    headers = {"Authorization": "Bearer static-token"}

    res = testclient.post_json("/oauth/register", payload, headers=headers, status=201)

    client = backend.get(models.Client, client_id=res.json["client_id"])
    assert res.json == {
        "client_id": client.client_id,
        "client_secret": client.client_secret,
        "client_id_issued_at": mock.ANY,
        **payload,
    }

    client = backend.get(models.Client, client_id=res.json["client_id"])
    for key, payload_value in payload.items():
        client_value = getattr(client, key)

        if isinstance(client_value, datetime):
            client_value = client_value.timestamp()

        if key == "scope":
            client_value = " ".join(client_value)

        assert client_value == payload_value
    backend.delete(client)



from canaille.app import models

from . import client_credentials


def test_nominal_case(testclient, client, keypair, trusted_client, backend, caplog):
    res = testclient.post(
        "/oauth/token",
        params=dict(
            grant_type="client_credentials",
            scope="openid profile email groups address phone",
        ),
        headers={"Authorization": f"Basic {client_credentials(client)}"},
        status=200,
    )

    access_token = res.json["access_token"]
    token = backend.get(models.Token, access_token=access_token)
    assert token.client == client
    assert token.subject is None
    assert set(token.scope) == {
        "openid",
        "profile",
        "email",
        "groups",
        "address",
        "phone",
    }



from canaille.app import models
from canaille.backends.ldap.backend import setup_ldap_models
from canaille.backends.ldap.ldapobject import LDAPObject


def test_guess_object_from_dn(backend, testclient, foo_group):
    foo_group.members = [foo_group]
    backend.save(foo_group)
    dn = foo_group.dn
    g = backend.get(LDAPObject, dn)
    assert isinstance(g, models.Group)
    assert g == foo_group
    assert g.display_name == foo_group.display_name


def test_object_class_update(backend, testclient):
    testclient.app.config["CANAILLE_LDAP"]["USER_CLASS"] = ["inetOrgPerson"]
    setup_ldap_models(testclient.app.config)

    user1 = models.User(cn="foo1", sn="bar1", user_name="baz1")
    backend.save(user1)

    assert set(user1.get_ldap_attribute("objectClass")) == {"inetOrgPerson"}
    assert set(
        backend.get(models.User, id=user1.id).get_ldap_attribute("objectClass")
    ) == {"inetOrgPerson"}

    testclient.app.config["CANAILLE_LDAP"]["USER_CLASS"] = [
        "inetOrgPerson",
        "extensibleObject",
    ]
    setup_ldap_models(testclient.app.config)

    user2 = models.User(cn="foo2", sn="bar2", user_name="baz2")
    backend.save(user2)

    assert set(user2.get_ldap_attribute("objectClass")) == {
        "inetOrgPerson",
        "extensibleObject",
    }
    assert set(
        backend.get(models.User, id=user2.id).get_ldap_attribute("objectClass")
    ) == {
        "inetOrgPerson",
        "extensibleObject",
    }

    user1 = backend.get(models.User, id=user1.id)
    assert user1.get_ldap_attribute("objectClass") == ["inetOrgPerson"]

    backend.save(user1)
    assert set(user1.get_ldap_attribute("objectClass")) == {
        "inetOrgPerson",
        "extensibleObject",
    }
    assert set(
        backend.get(models.User, id=user1.id).get_ldap_attribute("objectClass")
    ) == {
        "inetOrgPerson",
        "extensibleObject",
    }

    backend.delete(user1)
    backend.delete(user2)


def test_keep_old_object_classes(backend, testclient, slapd_server):
    """When using a populated LDAP database, some objects may have existing objectClass not handled by Canaille.

    In such a case Canaille should keep the unmanaged objectClass and
    attributes.
    """
    user = models.User(cn="foo", sn="bar", user_name="baz")
    backend.save(user)

    ldif = f"""dn: {user.dn}
changetype: modify
add: objectClass
objectClass: posixAccount
-
add: uidNumber
uidNumber: 1000
-
add: gidNumber
gidNumber: 1000
-
add: homeDirectory
homeDirectory: /home/foobar
"""

    process = slapd_server.ldapmodify(ldif)
    assert process.returncode == 0

    backend.reload(user)

    # saving an object should not raise a ldap.OBJECT_CLASS_VIOLATION exception
    backend.save(user)

    backend.delete(user)



import time
import uuid

from joserfc import jwt

from canaille.app import models

from . import client_credentials


def test_password_flow_basic(testclient, user, client, backend):
    res = testclient.post(
        "/oauth/token",
        params=dict(
            grant_type="password",
            username="user",
            password="correct horse battery staple",
            scope="openid profile groups",
        ),
        headers={"Authorization": f"Basic {client_credentials(client)}"},
        status=200,
    )

    assert res.json["scope"] == "openid profile groups"
    assert res.json["token_type"] == "Bearer"
    access_token = res.json["access_token"]

    token = backend.get(models.Token, access_token=access_token)
    assert token is not None

    res = testclient.get(
        "/oauth/userinfo",
        headers={"Authorization": f"Bearer {access_token}"},
        status=200,
    )
    assert res.json["name"] == "John (johnny) Doe"


def test_password_flow_post(testclient, user, client, backend):
    client.token_endpoint_auth_method = "client_secret_post"
    backend.save(client)

    res = testclient.post(
        "/oauth/token",
        params=dict(
            grant_type="password",
            username="user",
            password="correct horse battery staple",
            scope="openid profile groups",
            client_id=client.client_id,
            client_secret=client.client_secret,
        ),
        status=200,
    )

    assert res.json["scope"] == "openid profile groups"
    assert res.json["token_type"] == "Bearer"
    access_token = res.json["access_token"]

    token = backend.get(models.Token, access_token=access_token)
    assert token is not None

    res = testclient.get(
        "/oauth/userinfo",
        headers={"Authorization": f"Bearer {access_token}"},
        status=200,
    )
    assert res.json["name"] == "John (johnny) Doe"


def test_invalid_user(testclient, user, client):
    res = testclient.post(
        "/oauth/token",
        params=dict(
            grant_type="password",
            username="invalid",
            password="invalid",
            scope="openid profile groups",
        ),
        headers={"Authorization": f"Basic {client_credentials(client)}"},
        status=400,
    )

    assert res.json == {
        "error": "invalid_request",
        "error_description": "Invalid 'username' or 'password' in request.",
    }


def test_invalid_credentials(testclient, user, client):
    res = testclient.post(
        "/oauth/token",
        params=dict(
            grant_type="password",
            username="user",
            password="invalid",
            scope="openid profile groups",
        ),
        headers={"Authorization": f"Basic {client_credentials(client)}"},
        status=400,
    )

    assert res.json == {
        "error": "invalid_request",
        "error_description": "Invalid 'username' or 'password' in request.",
    }


def test_jwt_auth(testclient, client_jwks, client, token, caplog, backend, user):
    """Test client JWT authentication as defined per RFC7523, for a password grant."""
    now = time.time()

    client.trusted = True
    client.token_endpoint_auth_method = "client_secret_jwt"
    backend.save(client)

    _, private_key = client_jwks
    header = {"alg": "RS256"}
    payload = {
        "iss": client.client_id,
        "sub": client.client_id,
        "aud": "http://canaille.test/oauth/token",
        "nbf": now - 3600,
        "exp": now + 3600,
        "iat": now - 1,
        "jti": str(uuid.uuid4()),
    }
    client_jwt = jwt.encode(header, payload, private_key)
    res = testclient.post(
        "/oauth/token",
        params=dict(
            grant_type="password",
            username="user",
            password="correct horse battery staple",
            scope="openid profile groups",
            client_assertion=client_jwt,
            client_assertion_type="urn:ietf:params:oauth:client-assertion-type:jwt-bearer",
        ),
        status=200,
    )

    assert res.json["scope"] == "openid profile groups"
    assert res.json["token_type"] == "Bearer"
    access_token = res.json["access_token"]

    token = backend.get(models.Token, access_token=access_token)
    assert token is not None



import json
from unittest import mock

from canaille.commands import cli


def test_set_string_by_id(testclient, backend, user):
    """Set a string attribute to a model identifier by its id."""
    runner = testclient.app.test_cli_runner()
    res = runner.invoke(cli, ["set", "user", user.id, "--given-name", "foobar"])
    assert res.exit_code == 0, res.stdout
    assert json.loads(res.stdout) == {
        "created": mock.ANY,
        "display_name": "Johnny",
        "emails": [
            "john@doe.test",
        ],
        "family_name": "Doe",
        "formatted_address": "1235, somewhere",
        "formatted_name": "John (johnny) Doe",
        "given_name": "foobar",
        "id": user.id,
        "last_modified": mock.ANY,
        "password": mock.ANY,
        "phone_numbers": [
            "555-000-000",
        ],
        "preferred_language": "en",
        "profile_url": "https://john.test",
        "user_name": "user",
    }
    backend.reload(user)
    assert user.given_name == "foobar"


def test_set_string_by_identifier(testclient, backend, user):
    """Set a string attribute to a model identifier by its identifier."""
    runner = testclient.app.test_cli_runner()
    res = runner.invoke(cli, ["set", "user", "user", "--given-name", "foobar"])
    assert res.exit_code == 0, res.stdout
    assert json.loads(res.stdout) == {
        "created": mock.ANY,
        "display_name": "Johnny",
        "emails": [
            "john@doe.test",
        ],
        "family_name": "Doe",
        "formatted_address": "1235, somewhere",
        "formatted_name": "John (johnny) Doe",
        "given_name": "foobar",
        "id": user.id,
        "last_modified": mock.ANY,
        "password": mock.ANY,
        "phone_numbers": [
            "555-000-000",
        ],
        "preferred_language": "en",
        "profile_url": "https://john.test",
        "user_name": "user",
    }
    backend.reload(user)
    assert user.given_name == "foobar"


def test_set_multiple(testclient, backend, user):
    """Test setting several emails."""
    runner = testclient.app.test_cli_runner()
    res = runner.invoke(
        cli,
        [
            "set",
            "user",
            user.id,
            "--emails",
            "foo@example.test",
            "--emails",
            "bar@example.test",
        ],
    )
    assert res.exit_code == 0, res.stdout
    assert json.loads(res.stdout) == {
        "created": mock.ANY,
        "display_name": "Johnny",
        "emails": [
            "foo@example.test",
            "bar@example.test",
        ],
        "family_name": "Doe",
        "formatted_address": "1235, somewhere",
        "formatted_name": "John (johnny) Doe",
        "given_name": "John",
        "id": user.id,
        "last_modified": mock.ANY,
        "password": mock.ANY,
        "phone_numbers": [
            "555-000-000",
        ],
        "preferred_language": "en",
        "profile_url": "https://john.test",
        "user_name": "user",
    }
    backend.reload(user)
    assert user.emails == [
        "foo@example.test",
        "bar@example.test",
    ]


def test_set_unknown_id(testclient, backend):
    """Error case for trying to set a value for an invalid object."""
    runner = testclient.app.test_cli_runner()
    res = runner.invoke(cli, ["set", "user", "invalid", "--given-name", "foobar"])
    assert res.exit_code == 1, res.stdout
    assert res.stdout == "Error: No user with id 'invalid'\n"


def test_set_remove_simple_attribute(testclient, backend, user, admin):
    """Test to remove a non multiple attribute."""
    assert user.formatted_address is not None

    runner = testclient.app.test_cli_runner()
    res = runner.invoke(cli, ["set", "user", user.id, "--formatted-address", ""])
    assert res.exit_code == 0, res.stdout
    assert json.loads(res.stdout) == {
        "created": mock.ANY,
        "display_name": "Johnny",
        "emails": [
            "john@doe.test",
        ],
        "family_name": "Doe",
        "formatted_name": "John (johnny) Doe",
        "given_name": "John",
        "id": user.id,
        "last_modified": mock.ANY,
        "password": mock.ANY,
        "phone_numbers": [
            "555-000-000",
        ],
        "preferred_language": "en",
        "profile_url": "https://john.test",
        "user_name": "user",
    }
    backend.reload(user)
    assert user.formatted_address is None


def test_set_remove_multiple_attribute(testclient, backend, user, admin, foo_group):
    """Test to remove a non multiple attribute."""
    foo_group.members = [user, admin]
    backend.save(foo_group)
    assert user.groups == [foo_group]

    runner = testclient.app.test_cli_runner()
    res = runner.invoke(cli, ["set", "user", user.id, "--groups", ""])
    assert res.exit_code == 0, res.stdout
    assert json.loads(res.stdout) == {
        "created": mock.ANY,
        "display_name": "Johnny",
        "emails": [
            "john@doe.test",
        ],
        "family_name": "Doe",
        "formatted_name": "John (johnny) Doe",
        "formatted_address": "1235, somewhere",
        "given_name": "John",
        "id": user.id,
        "last_modified": mock.ANY,
        "password": mock.ANY,
        "phone_numbers": [
            "555-000-000",
        ],
        "preferred_language": "en",
        "profile_url": "https://john.test",
        "user_name": "user",
    }
    backend.reload(user)
    assert user.groups == []



from flask import current_app

from canaille.app.i18n import gettext as _
from canaille.backends import Backend


def get_user_from_login(login=None):
    from canaille.app import models

    login_attributes = current_app.config["CANAILLE"]["LOGIN_ATTRIBUTES"]

    if isinstance(login_attributes, list):
        for attr_name in login_attributes:
            if user := Backend.instance.get(models.User, **{attr_name: login}):
                return user

    if isinstance(login_attributes, dict):
        for attr_name, template in login_attributes.items():
            login_value = current_app.jinja_env.from_string(template).render(
                login=login
            )
            if user := Backend.instance.get(models.User, **{attr_name: login_value}):
                return user

    return None


def login_placeholder():
    """Build a placeholder string for the login form, based on the attributes users can use to identify themselves."""
    default_values = {
        "formatted_name": _("John Doe"),
        "user_name": _("jdoe"),
        "emails": _("john.doe@example.com"),
    }
    placeholders = [
        default_values[attr_name]
        for attr_name in current_app.config["CANAILLE"]["LOGIN_ATTRIBUTES"]
        if default_values.get(attr_name)
    ]
    return _(" or ").join(placeholders)



from scim2_tester import Status
from scim2_tester import check_server


def test_scim_tester(scim_client):
    results = check_server(scim_client, raise_exceptions=True)
    assert all(result.status == Status.SUCCESS for result in results)
